"""update_obsdb.py

Create and/or update an obsdb and obsfiledb based on some books.
The config file could be of the form:

.. code-block:: yaml

    base_dir: path_to_base_directories. Can be a list or a single string.
    obsdb_cols:
      start_time: float
      stop_time: float
      n_samples: int
      telescope: str
      tube_slot: str
      type: str
      subtype: str

    obsdb: dummyobsdb.sqlite
    obsfiledb: dummyobsfiledb.sqlite
    lat_tube_list_file: path to yaml dict matching tubes and bands
    tolerate_stray_files: True
    skip_bad_books: True
    known_bad_books_file: path to \n-separated file listing bad books 
    extra_extra_files:
    - Z_bookbinder_log.txt
    extra_files:
    - M_index.yaml
    - M_book.yaml

"""

from sotodlib.core.metadata import ObsDb
from sotodlib.core import Context 
from sotodlib.site_pipeline import check_book
from sotodlib.io import load_book
import os
import glob
import yaml
import re
import numpy as np
import time
import argparse
import logging
from sotodlib.site_pipeline import util
from typing import Optional
from itertools import product

logger = util.init_logger('update_obsdb', 'update-obsdb: ')

def check_meta_type(bookpath: str):
    metapath = os.path.join(bookpath, "M_index.yaml")
    meta = yaml.safe_load(open(metapath, "rb"))
    if meta is None:
        return "empty"
    elif "type" not in meta:
        return "notype"
    else:
        return meta["type"]


def telescope_lookup(telescope: str):
    """
    Set a number of common queries given a telescope name

    Arguments
    ----------
    telescope : str
        Name of telescope in M_index

    """
    if telescope == "sat" or telescope == "satp1":
        return {"telescope": "satp1", "telescope_flavor": "sat",
                "tube_flavor": "mf", "detector_flavor": "tes"}
    elif telescope == "satp2":
        return {"telescope": "satp2","telescope_flavor": "sat",
                "tube_flavor": "uhf", "detector_flavor": "tes"}
    elif telescope == "satp3":
        return {"telescope": "satp3", "telescope_flavor": "sat",
                "tube_flavor": "mf", "detector_flavor": "tes"}
    elif telescope == "lat":
        return {"telescope": "lat", "telescope_flavor": "lat",
                "detector_flavor": "tes"}
    else:
        logger.error("unknown telescope type given by bookbinder")
        return {}


def main(config: str, 
        recency: float = None, 
        booktype: Optional[str] = "both",
        verbosity: Optional[int] = 2,
        overwrite: Optional[bool] = False,
        fastwalk: Optional[bool] = False):

    """
    Create or update an obsdb for observation or operations data.

    Arguments
    ----------
    config : str
        Path to config file
    recency : float
        How far back in time to look for databases, in days. If None, 
        goes back to the UNIX start date (default: None)
    booktype : str
        Look for observations or operations data or both (default: both)
    verbosity : int
        Output verbosity. 0:Error, 1:Warning, 2:Info(default), 3:Debug
    overwrite : bool
        if False, do not re-check existing entries
    fastwalk : bool
        if True, assume the directories have a structure /base_dir/obs|oper/\d{5}/...
        Then replace base_dir with only the directories where \d{5} is greater or 
        equal to recency.
    """
    if verbosity == 0:
        logger.setLevel(logging.ERROR)
    elif verbosity == 1:
        logger.setLevel(logging.WARNING)
    elif verbosity == 2:
        logger.setLevel(logging.INFO)
    elif verbosity == 3:
        logger.setLevel(logging.DEBUG)

    logger.info("Updating obsdb")
    bookcart = []

    if booktype not in ["obs", "oper", "both"]:
        logger.warning("Specified booktype inadapted to update_obsdb")
    
    if booktype == "both":
        accept_type = ["obs", "oper"]
    else:
        accept_type = [booktype]

    config_dict = yaml.safe_load(open(config, "r"))
    try:
        base_dir = config_dict["base_dir"]
    except KeyError:
        logger.error("No base directory base_dir specified in config file!")

    if "obsdb" in config_dict:
        bookcartobsdb = ObsDb(map_file=config_dict["obsdb"])
    else:
        logger.warning("No obsdb named in the configuration file")
        bookcartobsdb = ObsDb("obsdb.sqlite")
        
    if "obsdb_cols" in config_dict:
        col_list = []
        for col, typ in config_dict["obsdb_cols"].items():
            col_list.append(col+" "+typ)
        bookcartobsdb.add_obs_columns(col_list)
    if "skip_bad_books" not in config_dict:
        config_dict["skip_bad_books"] = False
    
    config_dict["known_bad_books"] = []
    if "known_bad_books_file" in config_dict:
        try:
            with open(config_dict["known_bad_books_file"], "r") as bbf:
                config_dict["known_bad_books"] = bbf.read().split("\n")
        except:
            raise IOError("Bad books file couldn't be read in")
        
    #How far back we should look
    tnow = time.time()
    if recency is not None:
        tback = tnow - recency*86400
    else:
        tback = 0 #Back to the UNIX Big Bang 
    
    existing = bookcartobsdb.query()["obs_id"]
    #Check if there are one or multiple base_dir specified
    if isinstance(base_dir,str):
        base_dir = [base_dir]
    if fastwalk:
        abv_tback = int(f"{int(tback):05}"[:5]) #Make sure we have at least five chars
        abv_tnow = int(f"{int(tnow):05}"[:5])
        abv_codes = np.arange(abv_tback, abv_tnow+1)
        #Build the combinations base_dir/booktype/\d{5}
        base_dir = [f"{os.path.join(x[0], x[1], str(x[2]))}" for x in product(base_dir, accept_type, abv_codes)]
        logger.info(f"Looking in the following directories only: {str(base_dir)}")

    for bd in base_dir:
        #Find folders that are book-like and recent
        for dirpath, _, _ in os.walk(bd):
            if os.path.exists(os.path.join(dirpath, "M_index.yaml")):
                _, book_id = os.path.split(dirpath)
                if book_id in existing and not overwrite:
                    continue
                if book_id in config_dict["known_bad_books"]:
                    logger.debug(f"{book_id} known to be bad, skipping it")
                    continue
                edit_time = os.path.getmtime(dirpath)
                if edit_time > tback:
                    #Looks like a book folder and edited recently enough
                    bookcart.append(dirpath)
    
    logger.info(f"Found {len(bookcart)} new books in {time.time()-tnow} s")
    #Check the books for the observations we want
    bad_book_counter = 0
    for bookpath in sorted(bookcart):
        if check_meta_type(bookpath) in accept_type:
            t1 = time.time()
            logger.info(f"Examining book at {bookpath}")
            try:
                #obsfiledb creation
                ok, obsfiledb_info = check_book.scan_book_dir(
                    bookpath, logger, config_dict, prep_obsfiledb=True)
                if not ok:
                    raise RuntimeError("check_book found fatal errors, not adding.")
                check_book.add_to_obsfiledb(
                    obsfiledb_info, logger, config_dict, overwrite=True)
                logger.info(f"Ran check_book in {time.time()-t1} s")
            except Exception as e:
                if config_dict["skip_bad_books"]:
                    _, book_id = os.path.split(bookpath)
                    config_dict["known_bad_books"].append(book_id)
                    logger.error(f"failed to add {bookpath}. There are now {len(config_dict['known_bad_books'])} known bad books.")
                    bad_book_counter +=1
                    if "known_bad_books_file" in config_dict:
                        with open(config_dict["known_bad_books_file"], "w") as bbf:
                            bbf.write("\n".join(config_dict["known_bad_books"]))
                    continue
                else:
                    raise e

            index = yaml.safe_load(open(os.path.join(bookpath, "M_index.yaml"), "rb"))
            obs_id = index.pop("book_id")
            tags = index.pop("tags")
            detsets = index.pop("detsets")

            if "obsdb_cols" in config_dict:
                very_clean = {col:index[col] for col in iter(config_dict["obsdb_cols"]) if col in index}
            else:
                col_list = []
                clean = {key:val for key, val in index.items() if val is not None}
                very_clean = {key:val for key, val in clean.items() if type(val) is not list}
                for key, val in very_clean.items():
                    col_list.append(key+" "+type(val).__name__)
                bookcartobsdb.add_obs_columns(col_list)

            #Adding info that should be there for all observations
            #Descriptive string columns
            try:
                telescope = index["telescope"]
                flavors = telescope_lookup(telescope)
                for flav in flavors:
                    bookcartobsdb.add_obs_columns([flav+" str"])
                    very_clean[flav] = flavors[flav]
                if telescope == "lat":
                   lat_tube_list = yaml.safe_load(
                       open(config_dict["lat_tube_list_file"], "rb")
                   )
                   tube_flavor = lat_tube_list[index["tube_slot"]]
                   bookcartobsdb.add_obs_columns("tube_flavor str")
                   very_clean["tube_flavor"] = tube_flavor

            except KeyError:
                logger.error("No telescope key in index file or error with lat_tube_list")
                very_clean["telescope_flavor"] = "unknown"
            
            #Stream_ids and wafers
            try:
                stream_ids = index["stream_ids"]
                bookcartobsdb.add_obs_columns(["wafer_count int"])
                very_clean["wafer_count"] = len(stream_ids)
                wafer_slots = index["wafer_slots"]
                if len(wafer_slots) < len(stream_ids):
                    logger.error("Missing info on some stream_ids")
                    continue
                bookcartobsdb.add_obs_columns(["wafer_slots_list str", "stream_ids_list str"])
                wafer_slots_list = ""
                stream_ids_list = ",".join(stream_ids)
                for slot in wafer_slots:
                    if slot["stream_id"] in stream_ids:
                        wafer_slots_list += slot["wafer_slot"]+","
                very_clean["wafer_slots_list"] = wafer_slots_list[:-1]#Eliminate last comma
                very_clean["stream_ids_list"] = stream_ids_list
            except KeyError:
                logger.error("Unable to find stream_ids or wafer slots")

            #Time
            try:
                start = index["start_time"]
                end = index["stop_time"] 
                bookcartobsdb.add_obs_columns(["timestamp float", "duration float"])
                very_clean["timestamp"] = start
                very_clean["duration"] = end - start
            except KeyError:
                logger.error("Incomplete timing information for obs_id {obs_id}")
            
            #SAT HWP
            if very_clean["telescope_flavor"] == "sat":
                try:
                    very_clean["hwp_freq_mean"] = index["hwp_freq_mean"]
                    very_clean["hwp_freq_stdev"] = index["hwp_freq_stdev"]
                    bookcartobsdb.add_obs_columns(["hwp_freq_mean float", 
                                                   "hwp_freq_stdev float"])
                except KeyError:
                    logger.error(f"No HWP frequency info for obs_id {obs_id}")

            #Scanning motion
            stream_file = os.path.join(bookpath,"*{}*.g3".format(stream_ids[0]))
            stream = load_book.load_book_file(stream_file, no_signal=True)

            for coor in ["az", "el"]:
                try:
                    coor_enc = stream.ancil[coor+"_enc"]
                    bookcartobsdb.add_obs_columns([f"{coor}_center float", 
                                                   f"{coor}_throw float"])
                    very_clean[f"{coor}_center"] = round(.5 * (coor_enc.max() + coor_enc.min()), 4)
                    very_clean[f"{coor}_throw"] = round(.5 * (coor_enc.max() - coor_enc.min()), 4)
                except KeyError:
                    logger.error(f"No {coor} pointing in some streams for obs_id {obs_id}")

            try:
                if very_clean["telescope_flavor"] == "sat":
                    bore_enc = stream.ancil["boresight_enc"]
                    very_clean["roll_center"] = round(-.5 * (bore_enc.max() + bore_enc.min()), 4)
                    very_clean["roll_throw"] = round(.5 * (bore_enc.max() - bore_enc.min()), 4)
                if very_clean["telescope_flavor"] == "lat":
                    el_enc = stream.ancil["el_enc"]
                    corot_enc = stream.ancil["corotator_enc"]
                    roll = el_enc - 60. - corot_enc
                    very_clean["roll_center"] = round(.5 * (roll.max() + roll.min()), 4)
                    very_clean["roll_throw"] = round(.5 * (roll.max() - roll.min()), 4)

                bookcartobsdb.add_obs_columns(["roll_center float", "roll_throw float"])
            except KeyError:
                logger.error(f"Unable to compute roll for obs_id {obs_id}")
                

            # Make sure no invalid tags before update.
            tags = [t.strip() for t in tags if t.strip() != '']

            bookcartobsdb.update_obs(obs_id, very_clean, tags=tags)
            logger.info(f"Finished {obs_id} in {time.time()-t1} s")
        else:
            bookcart.remove(bookpath)
    if bad_book_counter != 0:
        logger.error(f"Found {bad_book_counter} new bad books, There are now {len(config_dict['known_bad_books'])} known bad books.")
        raise(Exception)

def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument("--config", help="ObsDb, ObsfileDb configuration file", 
        type=str, required=True)
    parser.add_argument("--recency", default=None, type=float,
        help="Days to subtract from now to set as minimum ctime. If None, no minimum")
    parser.add_argument("--verbosity", default=2, type=int,
        help="Increase output verbosity. 0:Error, 1:Warning, 2:Info(default), 3:Debug")
    parser.add_argument("--booktype", default="both", type=str,
        help="Select book type to look for: obs, oper, both(default)")
    parser.add_argument("--overwrite", action="store_true",
        help="If true, writes over existing entries")
    parser.add_argument("--fastwalk", action="store_true",
        help="Assume known directory tree shape and speed up walkthrough")
    return parser



if __name__ == '__main__':
    parser = get_parser(parser=None)
    args = parser.parse_args()
    main(**vars(args))



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Test toast workflow tools.
"""

import os
import shutil
import glob
import re

import numpy as np
import astropy.units as u

from unittest import TestCase

# Import so3g before any other packages that import spt3g
import so3g

from ._helpers import create_outdir

try:
    # Import sotodlib toast module first, which sets global toast defaults
    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    import sotodlib.toast.workflows as wrk
    import toast
    from toast.observation import default_values as defaults

    toast_available = True
except ImportError:
    raise
    toast_available = False


class ToastWorkflowTest(TestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        if not toast_available:
            print("TOAST cannot be imported- skipping unit tests", flush=True)
            return
        world, procs, rank = toast.get_world()
        self.outdir = create_outdir(fixture_name, mpicomm=world)

    def test_workflow_config(self):
        if not toast_available:
            return
        world, procs, rank = toast.get_world()
        testdir = os.path.join(self.outdir, "config")
        if world is None or world.rank == 0:
            if os.path.isdir(testdir):
                shutil.rmtree(testdir)
            os.makedirs(testdir)
        if world is not None:
            world.barrier()

        operators = list()
        wrk.setup_simulate_atmosphere_signal(operators)

        job, config, otherargs, runargs = wrk.setup_job(
            operators=operators,
            opts={
                "sim_atmosphere.enable": True,
                "sim_atmosphere.xstep": "Quantity('10.0 m')",
            },
        )

        self.assertTrue(job.operators.sim_atmosphere.enabled)
        self.assertTrue(job.operators.sim_atmosphere.xstep == 10.0 * u.meter)



import numpy as np
import so3g
from pixell import enmap

from .. import core, tod_ops, coords


def make_map(tod,
             P=None, wcs_kernel=None,
             res=0.1 * coords.DEG,
             dsT=None, demodQ=None, demodU=None,
             cuts=None,
             det_weights=None, det_weights_demod=None,
             center_on=None, flip_gamma=True):
    """Generates maps of temperature and polarization from demodulated HWP
    data.  It is assumed that each detector contributes a T signal
    (which has been low-pass filtered to avoid the modulated signal)
    stored in dsT, as well as separate Q and U timestreams,
    corresponding to the cosine-phase (demodQ = Q cos 4 chi) and
    sine-phase (demodU = U sin 4 chi) response of the HWP.

    The demodQ and demodU signals are assumed to have been computed
    without regard for the polarization angle of the detector, nor the
    on-sky parallactic angle.  The impact of these is handled by the
    projection routines in this function.

    Parameters
    ----------
    tod : sotodlib.core.AxisManager
        An AxisManager object
    P : sotodlib.coords.pmat
        Projection Matrix to be used for mapmaking. If None, it is
        generated from `wcs_kernel` or `center_on`.
    wcs_kernel : enlib.wcs.WCS or None, optional
        The WCS object used to generate the output map.
        If None, a new WCS object with a Cartesian projection and a resolution of `res` will be created.
    center_on: None or str
        Name of point source. If specified, Source-centered map with a tangent projection will be made.
        If None, wcs_kernel will be used to generate a projection matrix.
    size: float or None
        If specified, trim source-centored map to a local square with `size`, in radian.
        If None, not trimming will be applied. Only valid when `centor_on` is specified.
    res : float, optional
        The resolution of the output map, in radian.        
    dsT : array-like or None, optional
        The input dsT timestream data. If None, the 'dsT' field of `tod` will be used.
    demodQ : array-like or None, optional
        The input demodulated Q timestream data. If None, the 'demodQ' field of `tod` will be used.
    demodU : array-like or None, optional
        The input demodulated U timestream data. If None, the 'demodU' field of `tod` will be used.
    cuts : RangesMatrix or None, optional
        A RangesMatrix that identifies samples that should be excluded from projection operations.
        If None, no cuts will be applied.
    det_weights : array-like or None, optional
        The detector weights to use in the map-making for the dsT timestream.
    det_weights_demod : array-like or None, optional
        The detector weights to use in the map-making for the demodulated Q and U timestreams.
        If both of `det_weights` and `det_weights_demod` are None, uniform detector weights will be used.
        If only one of two are provided, the other weight is provided by `det_weights` = 2 * `det_weights_demod`.
    flip_gamma : bool or None
        Defaults to True.  When constructing the pointing matrix,
        reflect the nominal focal plane polarization angles about zero
        (gamma -> -gamma).  If you pass in your own P, make sure it
        was constructed with hwp=True.

    Returns
    -------
    A dictionary which contains:
    'map' : enmap.ndmap
        map of temperature and polarization
    'weighted_map' : enmap.ndmap
        The inverse variance weighted map of temperature and polarization
    'weight' : enmap.ndmap
        The map of inverse variance weights used in the map-making process.

    """
    if dsT is None:
        dsT = tod['dsT']
    if demodQ is None:
        demodQ = tod['demodQ']
    if demodU is None:
        demodU = tod['demodU']

    if flip_gamma is None:
        flip_gamma = True

    if P is None:
        if center_on is None:
            if wcs_kernel is None:
                wcs_kernel = coords.get_wcs_kernel('car', 0, 0, res)
            P = coords.P.for_tod(
                tod=tod, wcs_kernel=wcs_kernel, cuts=cuts, comps='QU', hwp=flip_gamma)
        else:
            P, X = coords.planets.get_scan_P(tod, planet=center_on, res=res, hwp=flip_gamma)
            

    if det_weights is None:
        if det_weights_demod is None:
            det_weights_demod = np.ones(tod.dets.count, dtype='float32')
        det_weights = det_weights_demod * 2.
    else:
        if det_weights_demod is None:
            det_weights_demod = det_weights / 2.

    # T map and weight
    mT_weighted = P.to_map(
        tod=tod, signal=dsT, comps='T', det_weights=det_weights)
    wT = P.to_weights(tod, signal=dsT, comps='T', det_weights=det_weights)

    # Q/U maps and weights
    mQ_weighted = P.to_map(tod=tod, signal=demodQ, comps='QU',
                             det_weights=det_weights_demod)
    mU_weighted = P.to_map(tod=tod, signal=demodU, comps='QU',
                             det_weights=det_weights_demod)

    mQU_weighted = P.zeros()
    mQU_weighted[0][:] = mQ_weighted[0] - mU_weighted[1]
    mQU_weighted[1][:] = mQ_weighted[1] + mU_weighted[0]
    del mQ_weighted, mU_weighted

    wQU = P.to_weights(tod, signal=demodQ, comps='T',
                         det_weights=det_weights_demod)

    # combine mT_weighted and mQU_weighted into mTQU_weighted
    mTQU_weighted = P.zeros(super_shape=3)
    mTQU_weighted[0] = mT_weighted
    mTQU_weighted[1] = mQU_weighted[0]
    mTQU_weighted[2] = mQU_weighted[1]

    # combine wT and wQU into wTQU
    wTQU = enmap.zeros((3, 3) + P.geom.shape, wcs=P.geom.wcs)
    wTQU[0][0] = wT
    wTQU[1][1] = wQU
    wTQU[2][2] = wQU

    # remove weights
    mTQU = P.remove_weights(signal_map=mTQU_weighted,
                              weights_map=wTQU, comps='TQU')
    
    output = {'map': mTQU,
             'weighted_map': mTQU_weighted,
             'weight': wTQU}
    return output

def from_map(tod, signal_map, cuts=None, flip_gamma=True, wrap=False, modulated=False):
    """
    Generate simulated TOD with HWP from a given signal map.

    Args:
        tod : an axisManager object
        signal_map: pixell.enmap.ndmap containing (Tmap, Qmap, Umap) representing the signal.
        cuts (RangesMatrix, optional): Cuts to apply to the data. Default is None.
        flip_gamma (bool, optional): Whether to flip detector coordinate. If you use the HWP, keep it `True`. Default is True.
        wrap (bool, optional): Whether to wrap the simulated data. Default is False.
        modulated (bool, optional): If True, return modulated signal. If False, return the demodulated signal 
        (`dsT`, `demodQ`, and `demodU`). Default is False. 

    Returns:
        `modulate==False`: A tuple containing the TOD (np.array) of dsT, demodQ and demodU.
        `modulate==True` : The modulated TOD (np.array)
        
    """
    Tmap, Qmap, Umap = signal_map
    
    P = coords.P.for_tod(tod=tod, geom=signal_map.geometry, cuts=cuts, 
                         comps='QU', hwp=flip_gamma)
    dsT_sim = P.from_map(Tmap, comps='T')
    demodQ_sim = P.from_map(enmap.enmap([Qmap, Umap]), comps='QU')
    demodU_sim = P.from_map(enmap.enmap([Umap, -Qmap]), comps='QU')
    
    if modulated is False:
        if wrap:
            tod.wrap('dsT', dsT_sim, [(0, 'dets'), (1, 'samps')])
            tod.wrap('demodQ', demodQ_sim, [(0, 'dets'), (1, 'samps')])
            tod.wrap('demodU', demodU_sim, [(0, 'dets'), (1, 'samps')])
        return dsT_sim, demodQ_sim, demodU_sim
    else:
        assert 'hwp_angle' in tod._fields
        signal_sim = dsT_sim + demodQ_sim*np.cos(4*tod.hwp_angle) + demodU_sim*np.sin(4*tod.hwp_angle)
        if wrap:
            tod.wrap('signal', signal_sim, [(0, 'dets'), (1, 'samps')])
        return signal_sim


def rotate_focal_plane(tod, hwp=True):
    """ Interpret the boresight rotation effect as a focal plane rotation and update them accordingly.
    This applies constant rotation to focal plane. If hwp=True, rotations of gamma are reflected.
    This updates tod.boresight.roll and tod.focal_plane, in place.
    """
    sign = -1. if hwp else 1.
    avg_roll = np.average(tod.boresight.roll)
    tod.boresight.roll -= avg_roll
    fp = so3g.proj.quat.rotation_xieta(tod.focal_plane.xi, tod.focal_plane.eta, sign * tod.focal_plane.gamma)
    fp = so3g.proj.quat.euler(2, avg_roll) * fp
    xi, eta, gamma = so3g.proj.quat.decompose_xieta(fp)
    # boresight is float64 but signal and focal_plane are float32
    tod.focal_plane.xi = np.float32(xi)
    tod.focal_plane.eta = np.float32(eta)
    tod.focal_plane.gamma = sign * np.float32(gamma)


def rotate_demodQU(tod, sign=1, offset=0, radial=False, update_focal_plane=True):
    """
    Apply detectors' polarization angle calibration to the HWP demodulated Q and U timestreams to
    place all detectors' Q and U timestreams in a common telescope frame. This updates tod.demodQ
    and tod.demodU, in place.
    To get Qr Ur timestreams, run `rotate_demodQU(tod)` and then `rotate_demodQU(tod, radial=True)`.
    To restore the Q U timestreams, run `rotate_demodQU(tod, sign=-1, radial=True)`.
    

    Args:
        tod : an axisManager object
        update_focal_plane (bool, optional): Whether to update focal_plane.gamma angles consistent with new coordinate reference. 
            Make this True for polarization mapmaking using make_map.
        offset : float, optional
            The rotation angle in degrees to apply (default is 0).
        sign : int, optional
            A sign factor to control the direction of the rotation (default is +1).
        radial : bool, optional
            If True and the Q U timestreams in tod are in a common telescope flame, this function turns Q U
            into Qr Ur. 
            

    """
    if radial:
        theta_fp = np.arctan2(tod.focal_plane.xi, tod.focal_plane.eta)
        demodC = ((tod.demodQ + 1j*tod.demodU).T * np.exp( sign*(-2j*theta_fp) )).T
        if update_focal_plane:
            tod.focal_plane.gamma -= sign * theta_fp

    else:
        demodC = ((tod.demodQ + 1j*tod.demodU).T * np.exp( sign*(-2j*tod.focal_plane.gamma + 1j*np.deg2rad(offset)) )).T
        if update_focal_plane:
            tod.focal_plane.gamma *= 0
    tod.demodQ = demodC.real
    tod.demodU = demodC.imag
    del demodC




import numpy as np
import os
import time
import yaml
import argparse
from typing import Optional, List, Callable

from sotodlib import core, preprocess
from sotodlib.io.metadata import write_dataset
from sotodlib.site_pipeline import util, jobdb
from sotodlib.utils.procs_pool import get_exec_env
from sotodlib.core.metadata.loader import LoaderError


dtype = [
    ('dets:readout_id', '<U40'),
    ('tau_hwp', 'f4'),
    ('tau_hwp_error', 'f4'),
    ('AQ', 'f4'),
    ('AQ_error', 'f4'),
    ('AU', 'f4'),
    ('AU_error', 'f4'),
    ('redchi2s', 'f4'),
]


def run(
    logger,
    context_path,
    process_pipe,
    metadata_list,
    obs_id,
    n_split,
):
    """
    Function to run process pipeline for tau_hwp.
    This split detectors into `n_split` chunks and process serially.
    Returns obs_id and ResultSet.
    """
    res_arr = np.array([], dtype=dtype)
    rset = core.metadata.ResultSet.from_friend(res_arr)

    try:
        ctx = core.Context(context_path, metadata_list=metadata_list)
        meta = ctx.get_meta(obs_id)
        nper = int(np.ceil(meta.dets.count/n_split))
        logger.info(f'process {obs_id}, dets {meta.dets.count}')
        for i in range(n_split):
            dets = meta.dets.vals[i*nper:(i+1)*nper]
            meta_short = meta.restrict('dets', dets, in_place=False)
            tod = ctx.get_obs(meta_short)
            pipe = preprocess.Pipeline(process_pipe)
            proc = core.AxisManager(tod.dets, tod.samps)
            for process in pipe:
                logger.info(f'{process.name}, dets {tod.dets.count}')
                process.process(tod, proc)
                process.calc_and_save(tod, proc)
            for i in range(tod.dets.count):
                dic = {'dets:readout_id': tod.dets.vals[i]}
                dic.update({v[0]: tod.tau_hwp[v[0]][i] for v in dtype[1:]})
                rset.append(dic)
        assert len(rset) == meta.dets.count
        return obs_id, rset
    except (LoaderError, OSError) as e:
        logger.error(f'Failed to load data {obs_id} {e}')
        return obs_id, None
    except Exception as e:
        logger.error(f'Failed to process {obs_id} {e}')
        return obs_id, None


def _main(
    executor,
    as_completed_callable: Callable,
    context_path: str,
    process_pipe: dict,
    output_dir: str,
    metadata_list: Optional[List[str]] = 'all',
    verbosity: Optional[int] = 2,
    overwrite: Optional[bool] = False,
    tags: Optional[List[str]] = None,
    obs_id: Optional[List[str]] = None,
    n_split: Optional[int] = 10,
    nprocs: Optional[int] = 1,
    max_retry: Optional[int] = 3,
    stale: Optional[float] = 60.
):
    """
    Main function for making tau_hwp metadata

    Arguments
    ---------
    context_path: str
        Path to context file.
    process_pipe: str
        Dictionary of process pipeline config
    output_dir: str
        Path to the output directory
    metadata_list: str or list of str
        List of metadata labels to load
    verbosity: int
        0: Error, 1: Warning, 2: Info, 3: Debug
    overwrite: bool
        If true, overwrites existing entries in the database
    tags: List of str
        List of tag to use for quering observations
    obs_id: List of str
        List of obs-ids of particular observations that you want to run.
    n_split:
        Number of splits for the serial processing of a single observation.
    nprocs: int
        Number of processes to use. 1 at site computing.
    max_retry: int
        Maximum number to retry job.
    stale: float
        Jobs locked for longer than this time will be treated as stale
        and unlocked.
    """

    logger = util.init_logger(
        __name__, 'make_tau_hwp: ', verbosity=verbosity)

    ctx = core.Context(context_path, metadata_list=metadata_list)
    obs_ids = []
    for tag in tags:
        obslist = ctx.obsdb.query("subtype = 'cal'", tags=[tag])
        for obs in obslist:
            obs_ids.append(obs['obs_id'])

    db_path = os.path.join(output_dir, 'tau_hwp.sqlite')
    if os.path.exists(db_path):
        logger.info(f"Mapping {db_path} for the "
                    "archive index.")
        db = core.metadata.ManifestDb(db_path)
    else:
        logger.info(f"Creating {db_path} for the "
                    "archive index.")
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('obs:obs_id')
        scheme.add_data_field('dataset')
        db = core.metadata.ManifestDb(
            db_path,
            scheme=scheme
        )

    jclass = 'tau_hwp'
    jdb_path = os.path.join(output_dir, 'jobdb.sqlite')
    jdb = jobdb.JobManager(sqlite_file=jdb_path)

    for obs_id in obs_ids:
        jobs = jdb.get_jobs(jclass=jclass, tags={'obs_id': obs_id})
        if len(jobs) == 0:
            jdb.create_job(jclass, tags={'obs_id': obs_id})

    for j in jdb.get_jobs(jclass=jclass):
        if j.lock and (time.time() - j.lock) > stale:
            logger.info(f'Clearing old lock on {j}')
            jdb.unlock(j.id)

    jstate = 'all' if overwrite else 'open'
    to_do = jdb.get_jobs(jclass=jclass, jstate=jstate, locked=False)

    futures = []
    with jdb.locked(to_do, count=len(to_do)) as jobs:
        for job in jobs:
            job.mark_visited()
            futures.append(executor.submit(
                run,
                logger,
                context_path,
                process_pipe,
                metadata_list,
                obs_id=job.tags['obs_id'],
                n_split=n_split,
            ))
        for future in futures:
            obs_id, rset = future.result()
            for job in jobs:
                if job.tags['obs_id'] == obs_id:
                    break
            if rset is not None:
                try:
                    logger.info(f'saving {obs_id}...')
                    unix = obs_id.split('_')[1][:4]  # first 4 digits
                    h5_fn = f'tau_hwp_{unix}.h5'
                    h5_path = os.path.join(output_dir, h5_fn)
                    write_dataset(
                        data=rset,
                        filename=h5_path,
                        address=obs_id,
                        overwrite=True
                    )
                    db.add_entry(
                        {'obs:obs_id': obs_id, 'dataset': obs_id},
                        filename=h5_fn, replace=overwrite,
                    )
                    job.jstate = 'done'
                    continue
                except Exception as e:
                    logger.error(f'Failed to save {obs_id} {e}')
            if job.visit_count > max_retry:
                logger.error(f'Mark {obs_id} as failed.')
                job.jstate = 'failed'
            else:
                logger.error(f'Failed {obs_id}, try again later')


def main(config):
    with open(config, "r") as f:
        cfg = yaml.safe_load(f)
    rank, executor, as_completed_callable = get_exec_env(nprocs=cfg['nprocs'])
    if rank == 0:
        _main(
            executor=executor,
            as_completed_callable=as_completed_callable,
            **cfg
        )


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('config', help="Path to configuration yaml file.")
    return parser


if __name__ == '__main__':
    util.main_launcher(main, get_parser)



import unittest
import tempfile
import os

import numpy as np
import astropy.units as u
from sotodlib import core
import sotodlib.core.axisman_util as amutil
import so3g

## "temporary" fix to deal with scipy>1.8 changing the sparse setup
try:
    from scipy.sparse import csr_array
except ImportError:
    from scipy.sparse import csr_matrix as csr_array

class TestAxisManager(unittest.TestCase):

    # Basic behavior of each axis type.

    def test_100_index(self):
        a1 = np.zeros(100)
        a1[10] = 1.
        aman = core.AxisManager(core.IndexAxis('samps', len(a1)))
        aman.wrap('a1', a1, [(0, 'samps')])
        # Don't let people wrap the same field twice
        with self.assertRaises(ValueError):
            aman.wrap('a1', 2*a1, [(0, 'samps')])
        aman.restrict('samps', (10, 30))
        self.assertNotEqual(aman.a1[0], 0.)
        self.assertEqual(len(aman.a1), 20)

    def test_110_offset(self):
        a1 = np.zeros(100)
        # Place the mark at index 10, and offset 15 -- so the mark is
        # at absolute index 25.
        a1[10] = 1.
        aman = core.AxisManager(
            core.OffsetAxis('samps', len(a1), 15))
        aman.wrap('a1', a1, [(0, 'samps')])
        aman.restrict('samps', (25, 30))
        self.assertNotEqual(aman.a1[0], 0.)
        self.assertEqual(len(aman.a1), 5)

    def test_120_label(self):
        dets = ['det0', 'det1', 'det2']
        a1 = np.zeros(len(dets))
        a1[1] = 1.
        aman = core.AxisManager(core.LabelAxis('dets', dets))
        aman.wrap('a1', a1, [(0, 'dets')])
        aman.restrict('dets', ['det1'])
        self.assertNotEqual(aman.a1[0], 0.)

        # Don't let people use non string labels
        dets_int = [0, 1, 2]
        with self.assertRaises(TypeError):
            aman = core.AxisManager(core.LabelAxis('dets', dets_int))

    def test_130_not_inplace(self):
        a1 = np.zeros(100)
        a1[10] = 1.
        aman = core.AxisManager(core.IndexAxis('samps', len(a1)))
        aman.wrap('a1', a1, [(0, 'samps')])
        aman.wrap('a2', 1)

        # This should return a separate thing.
        rman = aman.restrict('samps', (10, 30), in_place=False)
        # self.assertNotEqual(aman.a1[0], 0.)
        self.assertEqual(len(aman.a1), 100)
        self.assertEqual(len(rman.a1), 20)
        self.assertNotEqual(aman.a1[10], 0.)
        self.assertNotEqual(rman.a1[0], 0.)

    def test_140_restrict_axes(self):
        dets = ['det0', 'det1', 'det2']
        a1 = np.zeros((len(dets), 100))
        a1[1, 10] = 1.
        aman = core.AxisManager(core.LabelAxis('dets', dets),
                                core.OffsetAxis('samps', a1.shape[1]))
        aman.wrap('a1', a1, [(0, 'dets'), (1, 'samps')])
        aman.wrap('a2', 1)

        r_axes = {'dets': core.LabelAxis('dets', dets[1:2]),
                  'samps': core.OffsetAxis('samps', 20, 10)}
        # Not-in-place...
        rman = aman.restrict_axes(r_axes, in_place=False)
        self.assertEqual(aman.a1.shape, (3, 100))
        self.assertEqual(rman.a1.shape, (1, 20))
        self.assertNotEqual(aman.a1[1, 10], 0.)
        self.assertNotEqual(rman.a1[0, 0], 0.)
        # In-place.
        aman.restrict_axes(r_axes, in_place=True)
        self.assertEqual(aman.a1.shape, (1, 20))
        self.assertNotEqual(aman.a1[0, 0], 0.)

    def test_150_wrap_new(self):
        dets = ['det0', 'det1', 'det2']
        a1 = np.zeros((len(dets), 100))
        a1[1, 10] = 1.
        aman = core.AxisManager(core.LabelAxis('dets', dets),
                                core.OffsetAxis('samps', a1.shape[1]))
        x = aman.wrap_new('x', shape=('dets', 'samps'))
        y = aman.wrap_new('y', shape=('dets', 'samps'), dtype='float32')
        self.assertEqual(aman.x.shape, aman.y.shape)
        if hasattr(so3g.proj.RangesMatrix, 'zeros'):
            # Jan 2021 -- some so3g might not have this method yet...
            f = aman.wrap_new('f', shape=('dets', 'samps'),
                              cls=so3g.proj.RangesMatrix.zeros)
            self.assertEqual(aman.x.shape, aman.f.shape)

    def test_160_scalars(self):
        aman = core.AxisManager(core.LabelAxis('dets', ['a', 'b']),
                                core.OffsetAxis('samps', 100))

        # Accept trivially promoted scalars
        aman.wrap('x', 12)
        aman.wrap('z', 'hello')

        # Check that numpy int/float types are unpacked.
        aman.wrap('a', np.int32(12))
        aman.wrap('b', np.float32(12.))
        aman.wrap('c', np.str_('twelve'))
        self.assertNotIsInstance(aman['a'], np.integer)
        self.assertNotIsInstance(aman['b'], np.floating)
        self.assertNotIsInstance(aman['c'], np.str_)

        # Don't let people wrap the same scalar twice
        with self.assertRaises(ValueError):
            aman.wrap('x', 13)

        # Don't just let people wrap any old thing.
        with self.assertRaises(AttributeError):
            aman.wrap('a_dict', {'a': 123})
        with self.assertRaises(ValueError):
            aman.wrap('square_root', 1j)

        # Make sure AxisManager with scalar can be copied
        aman_copy = aman.copy()
        self.assertEqual(aman['x'], aman_copy['x'])

    def test_170_concat(self):
        # AxisManagers with shape (2, 100) and (3, 100)...
        detsA = ['det0', 'det1']
        detsB = ['det2', 'det3', 'det4']
        nsamps = 100
        amanA = core.AxisManager(core.LabelAxis('dets', detsA),
                                 core.OffsetAxis('samps', nsamps))
        amanB = core.AxisManager(core.LabelAxis('dets', detsB),
                                 core.OffsetAxis('samps', nsamps))

        # Empty AxisManagers ...
        aman = core.AxisManager.concatenate([amanA, amanB], axis='dets')
        self.assertEqual(aman.dets.count, len(detsA) + len(detsB))

        # Concat arrays?
        amanA.wrap_new('signal', shape=('dets', 'samps'))
        amanB.wrap_new('signal', shape=('dets', 'samps'))
        aman = core.AxisManager.concatenate([amanA, amanB], axis='dets')
        self.assertEqual(aman.signal.shape[0], len(detsA) + len(detsB))

        # Even if one is empty?
        amanX = amanA.restrict('dets', [])
        amanY = amanB.copy()
        aman = core.AxisManager.concatenate([amanX, amanY])
        self.assertEqual(aman.signal.shape[0], amanY.signal.shape[0])

        # or both are empty?
        amanY = amanB.restrict('dets', [])
        aman = core.AxisManager.concatenate([amanX, amanY])
        self.assertEqual(aman.signal.shape[0], 0)

        # or with sparse arrays?
        amanAA = core.AxisManager(core.LabelAxis('dets', detsA),
                                 core.OffsetAxis('samps', nsamps))
        amanBB = core.AxisManager(core.LabelAxis('dets', detsB),
                                 core.OffsetAxis('samps', nsamps))
        amanAA.wrap('sparse', csr_array( ((8,3), ([0,1], [1,21])), 
                                      shape=(amanAA.dets.count, amanAA.samps.count)),
                   [(0,'dets'),(1,'samps')])
        amanBB.wrap('sparse', csr_array( ((8,3), ([0,1], [2,54])), 
                                      shape=(amanBB.dets.count, amanBB.samps.count)),
                    [(0,'dets'),(1,'samps')])
        aman = core.AxisManager.concatenate([amanAA, amanBB])
        self.assertEqual(aman.dets.count, len(detsA) + len(detsB))

        # Handling of arrays++ that do not share the axis?
        for t in [amanA, amanB]:
            # Vectors, with nan.
            t.wrap_new('azimuth', shape=('samps',))[:] = 1.
            t.azimuth[3] = float('nan')

            # Scalars
            t.wrap("ans", 42)
            t.wrap("nans", float('nan'))

            # Include strings, because array_equal is picky on that...
            t.wrap('tersely', 'yo')
            t.wrap_new('greetings', shape=('samps',), dtype='S4')[:] = b'hi'
            t.wrap_new('free_greetings', shape=(4, ), dtype='U')[:] = 'hello'

            # And RangesMatrix
            t.wrap('rmat', so3g.proj.RangesMatrix.ones((5, 4)))


        amanB_backup, amanB = amanB, None

        # ... other_fields="exact"
        amanB = amanB_backup.copy()
        aman = core.AxisManager.concatenate([amanA, amanB], axis='dets')

        # ... other_fields="exact"
        amanB = amanB_backup.copy()
        amanB.azimuth[:] = 2.
        with self.assertRaises(ValueError):
            aman = core.AxisManager.concatenate([amanA, amanB], axis='dets')

        # RangesMatrix fail?
        amanB = amanB_backup.copy()
        amanB.rmat = so3g.proj.RangesMatrix.zeros(amanA.rmat.shape)
        with self.assertRaises(ValueError):
            aman = core.AxisManager.concatenate([amanA, amanB], axis='dets')

        # ... other_fields="exact" and arrays of different shapes
        amanB = amanB_backup.copy()
        amanB.move("azimuth", None)
        amanB.wrap("azimuth", np.array([43,5,2,3]))
        with self.assertRaises(ValueError):
            aman = core.AxisManager.concatenate([amanA, amanB], axis='dets')

        # ... other_fields="fail"
        amanB = amanB_backup.copy()
        amanB.move("azimuth",None)
        amanB.wrap_new('azimuth', shape=('samps',))[:] = 2.
        with self.assertRaises(ValueError):
            aman = core.AxisManager.concatenate([amanA, amanB], axis='dets',
                                               other_fields='fail')
        amanB.azimuth[:] = 1.
        with self.assertRaises(ValueError):
            aman = core.AxisManager.concatenate([amanA, amanB], axis='dets',
                                               other_fields='fail')

        # ... other_fields="drop"
        amanB = amanB_backup.copy()
        amanB.azimuth[:] = 2.
        aman = core.AxisManager.concatenate([amanA, amanB], axis='dets',
                                            other_fields="drop")
        self.assertNotIn('azimuth', aman)

        # ... other_fields="first"
        aman = core.AxisManager.concatenate([amanA, amanB], axis='dets',
                                            other_fields='first')
        self.assertIn('azimuth', aman)
        self.assertSequenceEqual(aman._assignments['azimuth'], ['samps'])
        self.assertEqual(aman.azimuth[0], 1.)

        # Handling of child AxisManagers
        fpA = core.AxisManager(amanA.dets.copy())
        fpB = core.AxisManager(amanB.dets.copy())
        fpA.wrap_new('x', shape=('dets',))
        fpB.wrap_new('x', shape=('dets',))
        amanA.wrap('fp', fpA)
        amanB.wrap('fp', fpB)
        aman = core.AxisManager.concatenate([amanA, amanB], axis='dets',
                                            other_fields='drop')
        self.assertSequenceEqual(aman.fp.shape, (aman.dets.count, ))

        # Loop checking
        a = amanA.copy()
        b = amanB.copy()
        a.wrap('b', b)
        with self.assertRaises(AssertionError):
            a.wrap('a', a)
        with self.assertRaises(AssertionError):
            a.b.wrap('a', a)
        # This is allowed because a.b is not the same as b.  Maybe it
        # should be... but that will be a deliberate API change.
        b.wrap('a', a)
        self.assertNotIn('a', a.b)

    def test_180_overwrite(self):
        dets = ['det0', 'det1', 'det2']
        a1 = np.zeros((len(dets), 100))
        a1[1, 10] = 1.
        aman = core.AxisManager(core.LabelAxis('dets', dets),
                                core.OffsetAxis('samps', a1.shape[1]))
        aman.wrap('a1', a1, [(0, 'dets'), (1, 'samps')])
        a2 = np.zeros((len(dets), 100))
        a2[2, 11] = 1.
        aman.wrap('a1', a2, [(0, 'dets'), (1, 'samps')],
                  overwrite=True)
        self.assertNotEqual(aman.a1[2,11], 0)
        self.assertNotEqual(aman.a1[1,10], 1.)

    def test_190_get_set(self):
        dets = ["det0", "det1", "det2"]
        n, ofs = 1000, 0
        aman = core.AxisManager(
            core.LabelAxis("dets", dets), core.OffsetAxis("samps", n, ofs)
        )
        child = core.AxisManager(
            core.LabelAxis("dets", dets + ["det3"]),
            core.OffsetAxis("samps", n, ofs - n // 2),
        )

        child2 = core.AxisManager(
            core.LabelAxis("dets2", ["det4", "det5"]),
            core.OffsetAxis("samps", n, ofs - n // 2),
        )
        child2.wrap("tod", np.zeros((2, 1000)))
        aman.wrap("child", child)
        aman["child"].wrap("child2", child2)
        self.assertEqual(aman["child.child2.dets2"].count, 2)
        self.assertEqual(aman["child.dets"].name, "dets")
        np.testing.assert_array_equal(
            aman["child.child2.dets2"].vals, np.array(["det4", "det5"])
        )
        self.assertEqual(aman["child.child2.samps"].count, n // 2)
        self.assertEqual(aman["child.child2.samps"].offset, 0)
        self.assertEqual(
            aman["child.child2.samps"].count, aman.child.child2.samps.count
        )
        self.assertEqual(
            aman["child.child2.samps"].offset, aman.child.child2.samps.offset
        )

        np.testing.assert_array_equal(aman["child.child2.tod"], np.zeros((2, 1000)))

        with self.assertRaises(KeyError):
            aman["child2"]

        with self.assertRaises(AttributeError):
            aman["child.dets.an_extra_layer"]

        self.assertIn("child.dets", aman)
        self.assertIn("child.dets2", aman)  # I am not sure why this is true
        self.assertNotIn("child.child2.someentry", aman)
        self.assertNotIn("child.child2.someentry.someotherentry", aman)

        with self.assertRaises(ValueError):
            aman["child"] = child2

        new_tods = np.ones((2, 500))
        aman.child.child2.tod = new_tods
        np.testing.assert_array_equal(aman["child.child2.tod"], np.ones((2, 500)))
        np.testing.assert_array_equal(aman.child.child2.tod, np.ones((2, 500)))

        new_tods = np.ones((2, 1500))
        aman["child.child2.tod"] = new_tods
        np.testing.assert_array_equal(aman["child.child2.tod"], np.ones((2, 1500)))
        np.testing.assert_array_equal(aman.child.child2.tod, np.ones((2, 1500)))

        aman.move('child.child2.tod', 'child.child2.tod_new')
        aman.move('child.child2.tod_new', 'child.tod')
        aman.move('child.child2', None)

    # Multi-dimensional restrictions.

    def test_200_multid(self):
        dets = ['det0', 'det1', 'det2']
        a1 = np.zeros((len(dets), len(dets)))
        a1[2, 2] = 1.
        aman = core.AxisManager(core.LabelAxis('dets', dets))
        aman.wrap('a1', a1, [(0, 'dets'), (1, 'dets')])
        aman.restrict('dets', ['det1', 'det2'])
        self.assertEqual(aman.a1.shape, (2, 2))
        self.assertNotEqual(aman.a1[1, 1], 0.)

    def test_300_restrict(self):
        dets = ['det0', 'det1', 'det2']
        n, ofs = 1000, 5000
        aman = core.AxisManager(
            core.LabelAxis('dets', dets),
            core.OffsetAxis('samps', n, ofs))
        # Super-correlation matrix.
        a1 = np.zeros((len(dets), len(dets), n, n))
        a1[1, 1, 20, 21] = 1.
        aman.wrap('a1', a1, [(0, 'dets'), (1, 'dets'),
                             (2, 'samps'), (3, 'samps')])
        aman.restrict('dets', ['det1']).restrict('samps', (20 + ofs, 30 + ofs))
        self.assertEqual(aman.shape, (1, 10))
        self.assertEqual(aman.a1.shape, (1, 1, 10, 10))
        self.assertNotEqual(aman.a1[0, 0, 0, 1], 0.)

    # wrap of AxisManager, merge.

    def test_400_child(self):
        dets = ['det0', 'det1', 'det2']
        n, ofs = 1000, 0
        aman = core.AxisManager(
            core.LabelAxis('dets', dets),
            core.OffsetAxis('samps', n, ofs))
        child = core.AxisManager(
            core.LabelAxis('dets', dets + ['det3']),
            core.OffsetAxis('samps', n, ofs - n//2))
        aman.wrap('child', child)
        self.assertEqual(aman.shape, (3, n//2))
        self.assertEqual(aman._axes['samps'].offset, ofs)
        # See that aman.child and child are different...
        self.assertEqual(aman.child.dets.count, 3)
        self.assertEqual(child.dets.count, 4)

        # But you can tell it it's ok.
        child2 = core.AxisManager(
            core.LabelAxis('dets', dets + ['det3']),
            core.OffsetAxis('samps', n, ofs - n//2))
        aman.wrap('child2', child2, restrict_in_place=True)
        self.assertEqual(aman.child2.dets.count, 3)
        self.assertEqual(child2.dets.count, 3)

    def test_401_restrict(self):
        # Test AxisManager.restrict when it has AxisManager members.
        dets = ['det0', 'det1', 'det2']
        n, ofs = 1000, 0
        for in_place in [True, False]:
            aman = core.AxisManager(
                core.LabelAxis('dets', dets),
                core.OffsetAxis('samps', n, ofs))
            child = core.AxisManager(aman.dets, aman.samps)
            child2 = core.AxisManager(
                core.LabelAxis('not_dets', ['x', 'y', 'z']))
            aman.wrap('child', child)
            aman.wrap('rebel_child', child2)
            aout = aman.restrict('dets', ['det1'], in_place=in_place)
            msg = f'Note restrict was with in_place={in_place}'
            self.assertTrue(aout is aman or not in_place, msg=msg)
            self.assertEqual(aout['child'].shape, (1, n), msg=msg)
            self.assertIn('rebel_child', aout, msg=msg)
            self.assertEqual(aout['rebel_child'].shape, (3,), msg=msg)

    def test_402_restrict_axes(self):
        # Test AxisManager.restrict_axes when it has AxisManager members.
        dets = ['det0', 'det1', 'det2']
        n, ofs = 1000, 0
        for in_place in [True, False]:
            aman = core.AxisManager(
                core.LabelAxis('dets', dets),
                core.OffsetAxis('samps', n, ofs))
            child = core.AxisManager(aman.dets, aman.samps)
            child2 = core.AxisManager(
                core.LabelAxis('not_dets', ['x', 'y', 'z']))
            aman.wrap('child', child)
            aman.wrap('rebel_child', child2)
            new_dets = core.LabelAxis('dets', ['det1'])
            aout = aman.restrict_axes([new_dets], in_place=in_place)
            msg = f'Note restrict was with in_place={in_place}'
            self.assertTrue(aout is aman or not in_place, msg=msg)
            self.assertEqual(aout['child'].shape, (1, n), msg=msg)
            self.assertIn('rebel_child', aout, msg=msg)
            self.assertEqual(aout['rebel_child'].shape, (3,), msg=msg)

    def test_403_restrict_slices(self):
        # Test in_place is false for slices of arrays
        dets = ['det0', 'det1', 'det2']
        n, ofs = 1000, 0
        aman = core.AxisManager(
                core.LabelAxis('dets', dets),
                core.OffsetAxis('samps', n, ofs)
        )
        aman.wrap_new('test', ('dets','samps'))
        rman = aman.restrict('samps', (0,500), in_place=False)
        rman.test += 5
        self.assertNotEqual( aman.test[0,0], rman.test[0,0])

    def test_410_merge(self):
        dets = ['det0', 'det1', 'det2']
        n, ofs = 1000, 0
        for in_place in [False, True]:
            aman = core.AxisManager(
                core.LabelAxis('dets', dets),
                core.OffsetAxis('samps', n, ofs))
            coparent = core.AxisManager(
                core.LabelAxis('dets', dets + ['det3']),
                core.OffsetAxis('samps', n, ofs - n//2))\
                .wrap('x', np.arange(n), [(0, 'samps')])
            if in_place:
                kw = {'restrict_in_place': True}
            else:
                kw = {}
            aman.merge(coparent, **kw)
            self.assertEqual(aman.shape, (3, n//2))
            self.assertEqual(aman._axes['samps'].offset, ofs)
            self.assertEqual(aman.x[0], n//2)

            if in_place:
                # Check coparent was modified
                self.assertEqual(coparent.shape, aman.shape)
            else:
                # Check coparent was not modified in place
                self.assertEqual(coparent.shape, (4, n))


    def test_500_io(self):
        # Test save/load HDF5
        dets = ['det0', 'det1', 'det2']
        n, ofs = 1000, 0
        aman = core.AxisManager(
            core.LabelAxis('dets', dets),
            core.OffsetAxis('samps', n, ofs),
            core.IndexAxis('indexaxis', 12))
        # Make sure this has axes, scalars, a string array ...
        aman.wrap_new('test1', ('dets', 'samps'), dtype='float32')
        aman.wrap_new('flag1', shape=('dets', 'samps'),
                      cls=so3g.proj.RangesMatrix.zeros)
        aman.wrap('scalar', 8)
        aman.wrap('test_str', np.array(['a', 'b', 'cd']))
        aman.wrap('flags', core.FlagManager.for_tod(aman, 'dets', 'samps'))

        aman.wrap('a', np.int32(12))
        aman.wrap('b', np.float32(12.))
        aman.wrap('c', np.str_('twelve'))
        aman.wrap('d', np.bool_(False))

        aman.wrap('rangesint32', so3g.proj.Ranges.from_array(
                  np.array([[0, 1], [10, 20]]).astype(np.int32), aman.samps.count))
        aman.wrap('sparse', csr_array( ((8,3), ([0,1], [1,54])), 
                                      shape=(aman.dets.count, aman.samps.count)))

        aman.wrap('quantity', np.ones(5) << u.m)
        aman.wrap('quantity2', (np.ones(1) << u.m)[0])

        aman.wrap('quantized', np.sin(np.linspace(0, 6, 10001)).round(2))
        aman.wrap('quantized2', (aman['quantized'] * 100).astype('int32'))

        aman.wrap('subaman', core.AxisManager(
            aman.dets, core.LabelAxis('bands', ['f090', 'f150'])))
        aman['subaman'].wrap_new('subtest1', shape=[('dets', 'bands')])
        aman['subaman'].wrap_new('subtest2', shape=(100,))
        aman['subaman'].wrap('also_q', np.cos(np.linspace(0, 6, 10001)))

        encodings = {'quantized': {'type': 'flacarray', 'args': {'quanta': .01}},
                     'quantized2': {'type': 'flacarray'},
                     'subaman': {'also_q': {'type': 'flacarray', 'args': {'quanta': .2}}}}

        # Make sure the saving / clobbering / readback logic works
        # equally for simple group name, root group, None->root group.
        for dataset in ['path/to/my_axisman', '/', None]:
            with tempfile.TemporaryDirectory() as tempdir:
                filename = os.path.join(tempdir, 'test.h5')
                aman.save(filename, dataset)
                # Overwrite
                aman.save(filename, dataset, overwrite=True)
                # Fail on encodings error.
                with self.assertRaises(ValueError):
                    aman.save(filename, dataset, overwrite=True,
                              encodings={'something': {'type': 'whatever'}})
                # Fail on encodings error.
                with self.assertRaises(ValueError):
                    aman.save(filename, dataset, overwrite=True,
                              encodings={'quantity': {'type': 'flacarray'}})
                # Compress and Overwrite
                aman.save(filename, dataset, overwrite=True, compression='gzip',
                          encodings=encodings)
                # Refuse to overwrite
                with self.assertRaises(RuntimeError):
                    aman.save(filename, dataset)
                # Read back.
                aman2 = aman.load(filename, dataset)
            # Check what was read back.
            # This is not a very satisfying comparison ... support for ==
            # should be required for all AxisManager members!
            for k in aman._fields.keys():
                self.assertEqual(aman[k].__class__, aman2[k].__class__)
                if isinstance(aman[k], u.quantity.Quantity):
                    np.testing.assert_array_equal(aman[k], aman2[k])
                elif isinstance(aman[k], np.ndarray):
                    if np.issubdtype(aman2[k].dtype, np.floating):
                        np.testing.assert_almost_equal(aman[k], aman2[k])
                    else:
                        np.testing.assert_array_equal(aman[k], aman2[k])
                elif hasattr(aman[k], 'shape'):
                    self.assertEqual(aman[k].shape, aman2[k].shape)
                else:
                    self.assertEqual(aman[k], aman2[k])  # scalar

        # Test field subset load
        with tempfile.TemporaryDirectory() as tempdir:
            filename = os.path.join(tempdir, 'test.h5')
            aman.save(filename, dataset, encodings=encodings)
            aman2 = aman.load(filename, dataset, fields=['test1', 'flags'])
            aman3 = aman.load(filename, dataset, fields=['test1', 'subaman'])
            aman4 = aman.load(filename, dataset, fields=['test1', 'subaman.subtest1'])
        for target, keys, subkeys in [
                (aman2, ['test1', 'flags'], None),
                (aman3, ['test1', 'subaman'], ['subtest1', 'subtest2', 'also_q']),
                (aman4, ['test1', 'subaman'], ['subtest1']),
        ]:
            self.assertCountEqual(target._fields.keys(),
                                  keys)
            if subkeys:
                self.assertCountEqual(target['subaman']._fields.keys(),
                                      subkeys)

    def test_900_everything(self):
        tod = core.AxisManager(
            core.LabelAxis('dets', list('abcdef')),
            core.OffsetAxis('samps', 1000))
        cal = core.AxisManager(
            core.LabelAxis('dets', list('feghij')))
        cuts = core.AxisManager(
            core.OffsetAxis('samps', 800, 100))
        tod.wrap('data', np.ones(tod.shape, 'float32'), )
        cal.wrap('cal', np.linspace(.9, 1.2, 6), [(0, 'dets')])
        cuts.wrap('cuts', np.ones(cuts.shape, 'int32'), [(0, 'samps')])
        tod.merge(cal, cuts)
        self.assertEqual(tod.shape, (2, 800))

class TestFlagManager(unittest.TestCase):
    
    def test_100_inheritance(self):
        tod = core.AxisManager(
            core.LabelAxis('dets', list('abcdef')),
            core.OffsetAxis('samps', 1000))
        flags = core.FlagManager.for_tod(tod, 'dets', 'samps')
        tod.wrap('flags', flags)
        self.assertTrue( type(tod.flags) == core.FlagManager )


class TestUtil(unittest.TestCase):
    def test_coindices(self):
        x = ['d', 'a', 'g', 'e', 'b']
        y = ['x', 'b', 'o', 'd', 'q']
        z, i0, i1 = core.util.get_coindices(x, y)
        the_answers = (['d', 'b'], [0, 4], [3, 1])
        self.assertEqual(list(z),  the_answers[0])
        self.assertEqual(list(i0), the_answers[1])
        self.assertEqual(list(i1), the_answers[2])

    def test_multi_index(self):
        x = ['x', 'y', 'c', 'a']
        y = ['a', 'a', 'b', 'c', 'a', 'u']
        ix = core.util.get_multi_index(x, y)
        the_answer = np.array([x.index(_y) if _y in x else -1
                               for _y in y])
        self.assertEqual(list(ix), list(the_answer))


class TestAxisManagerUtil(unittest.TestCase):
    def test_restrict_times(self):
        am = core.AxisManager()
        nsamps = 100
        am.wrap('timestamps', np.arange(nsamps),
                [(0, core.OffsetAxis('samps', nsamps))])
        am2 = amutil.restrict_to_times(am, 10, 20)
        self.assertEqual(len(am2.timestamps),10)

    def test_restrict_times_raises(self):
        am = core.AxisManager()
        nsamps = 100
        am.wrap('timestamps', np.arange(nsamps),
                [(0, core.OffsetAxis('samps', nsamps))])
        with self.assertRaises(amutil.RestrictionException):
            amutil.restrict_to_times(am, 200, 300)

    def test_dict_to_am(self):
        d = {
            'hi': 10,
            'nested': {
                'abcd': np.array([0, 1]),
                'test': 'answer',
            }
        }
        am = amutil.dict_to_am(d)
        self.assertEqual(am.nested.test, 'answer')


class TestResultSet(unittest.TestCase):
    def test_00_basic(self):
        keys = ['A', 'B', 'C']
        dtypes = ['int64', 'float64', 'unicode']
        rows = [
            (1, 1., 'a'),
            (2, 2., 'b'),
            (3, 3., 'c'),
            (4, 4., 'd'),
        ]

        # Test array conversion
        rs = core.metadata.ResultSet(keys=keys, src=[list(r) for r in rows])
        for k, dtype in zip(keys, dtypes):
            assert len(rs[k]) == len(rows)
            assert np.issubdtype(rs[k].dtype, dtype)

        rsa = rs.asarray()
        assert len(rsa) == len(rs)
        for k, dtype in zip(keys, dtypes):
            assert np.issubdtype(rsa.dtype[k], dtype)

        # Even with nulls
        for i in range(3):
            r = list(rs.rows[i])
            r[i] = None
            rs.rows[i] = tuple(r)

        for k, dtype in zip(keys, dtypes):
            assert len(rs[k]) == len(rows)
            assert np.issubdtype(rs[k].dtype, dtype)

        rsa = rs.asarray()
        assert len(rsa) == len(rs)
        for k, dtype in zip(keys, dtypes):
            assert np.issubdtype(rsa.dtype[k], dtype)

        # Subsetting
        rs = core.metadata.ResultSet(keys=keys, src=[list(r) for r in rows])

        rs1 = rs.subset(keys=['B', 'A'])
        assert rs1.rows[0] == (1., 1)

        rs2 = rs.subset(rows=rs['A'] == 1)
        assert (len(rs2) == 1)

        rs3 = rs.subset(rows=np.array([1, 3]))
        assert (len(rs3) == 2)

        # Conversion.
        aman = rs.to_axismanager(axis_key='C')

        # Merge
        rs2 = core.metadata.ResultSet(keys=['D'], src=[[1], [2], [3], [4]])
        rs.merge(rs2)
        assert (rs.keys == ['A', 'B', 'C', 'D'])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2018-2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simons Observatory I/O routines.

This module contains code for reading and writing data and metadata.

"""



import math
import os
import inspect
import logging
import time
import sys
import copy
import argparse
import yaml
import copy
import requests
import datetime as dt

class ArchivePolicy:
    """Storage policy assistance.  Helps to determine the HDF5
    filename and dataset name for a result.

    Make me better!

    """
    @staticmethod
    def from_params(params):
        if params['type'] == 'simple':
            return ArchivePolicy(**params)
        if params['type'] == 'directory':
            return DirectoryArchivePolicy(**params)
        raise ValueError('No handler for "type"="%s"' % params['type'])

    def __init__(self, **kwargs):
        self.filename = kwargs['filename']

    def get_dest(self, product_id):
        """Returns (hdf_filename, dataset_addr).

        """
        return self.filename, product_id

class ArgumentParser(argparse.ArgumentParser):
    """A variant of ArgumentParser that allows the defaults
    to be overriden by values in a yaml config files. Thus the
    priority order becomes, from highest to lowest:

    1. Arguments passed on the command line
    2. The config file
    3. Defaults defined with add_argument()

    The config file is specified using the --config-file option,
    which this class adds automatically. It should therefore not
    be added manually.
    """
    def __init__(self, *args, **kwargs):
        argparse.ArgumentParser.__init__(self, *args, **kwargs)
        self.add_argument("--config-file", type=str, default=None,
            help="Optional yaml file containing overrides for the default values")
    def parse_args(self, argv=None):
        parser = argparse.ArgumentParser()
        parser.add_argument("--config-file", type=str, default=None)
        args, _ = parser.parse_known_args()
        if args.config_file != None:
            # Ok, we have a config file, parse it and use it to
            # replace our defaults
            with open(args.config_file, "r") as ifile:
                config = yaml.safe_load(ifile)
            for action in self._actions:
                try:
                    action.default  = config[action.dest]
                    # We mark it as non-required so we can run
                    # without even normally required arguments
                    # if they're provided by the config file
                    action.required = False
                except (KeyError, AttributeError) as e:
                    pass
        # Then parse again, taking into account any default update
        return argparse.ArgumentParser.parse_args(self, argv)

class DirectoryArchivePolicy:
    """Storage policy for stuff organized directly on the filesystem.

    """
    def __init__(self, **kwargs):
        self.root_dir = kwargs['root_dir']
        self.pattern = kwargs['pattern']

    def get_dest(self, **kw):
        """Returns full path to destination directory.

        """
        return os.path.join(self.root_dir, self.pattern.format(**kw))


def parse_quantity(val, default_units=None):
    """Convert an expression with units into an astropy Quantity.

    Args:
      val: the expression (see Notes).
      default_units: the units to assume if they are not provided in
        val.

    Returns:
      The astropy Quantity decoded from the argument.  Note the
      quantity is converted to the default_units, if they are
      provided.

    Notes:
      The default_units, if provided, should be "unit-like", by which
      we mean it is either:

      - An astropy Unit.
      - A string that astropy.units.Unit() can parse.

      The val can be any of the following:

      - A tuple (x, u) or list [x, u], where x is a float and u is
        unit-like.
      - A string (x), where x can be parsed by astropy.units.Quantity.
      - A float (x), but only if default_units is not None.

    Examples:
      >>> parse_quantity('100 arcsec')
      <Quantity 100. arcsec>

      >>> parse_quantity([12., 'deg'])
      <Quantity 12. deg>

      >>> parse_quantity('15 arcmin', 'deg')
      <Quantity 0.25 deg>

      >>> parse_quantity(100, 'm')
      <Quantity 100. m>

    """
    # Heavy to import, and we want this module to be fast to import
    # because it provides an ArgumentParser that should inform us
    # of incorrect arguments with as low latency as possible
    from astropy import units as u

    if default_units is not None:
        default_units = u.Unit(default_units)

    if isinstance(val, str):
        q = u.Quantity(val)
    elif isinstance(val, (list, tuple)):
        q = val[0] * u.Unit(val[1])
    elif isinstance(val, (float, int)):
        if default_units is None:
            raise ValueError(
                f"Cannot decode argument '{val}' without default_units.")
        q = val * default_units

    if default_units is not None:
        q = q.to(default_units)
    return q


def _filter_dict(d, bad_keys=['_stop_here']):
    if not isinstance(d, dict):
        return d
    # Support for lookup_conditional
    return {k: v for k, v in d.items()
            if k not in bad_keys}


def lookup_conditional(source, key, tags=None, default=KeyError):
    """Lookup a value in a dict, with the possibility of descending
    through nested dictionaries using tags provided by the user.

    This function returns the returns source[key] unless source[key]
    is a dict, in which case the tags (a list of strings) are each
    tested in the dict to see if they lead to a sub-setting.

    For example, if the source dictionary is {'number': {'a': 1, 'b':
    2}} and the user requests key 'number', with tags=['a'], then the
    returned value will be 1.

    If you want a dict to be returned literally, and not crawled
    further, include a dummy key '_stop_here', with arbitrary value
    (this key will be removed from the result before returning to the
    user).

    The key '_default' will always cause a match, even if none of the
    other tags match.  (This _default value also becomes the default
    if further recursion fails to yield an exact match.)

    Args:
      source (dict): The parameter tree to search.
      key (str): The key to terminate the search on.
      tags (list of str or None): tags that may be auto-descended.
      default: Value to return if the search does not resolve.  The
        special value KeyError will instead cause a KeyError to be
        raised if the search is not resolved.

    Examples::

      source = {
        'my_param': {
          '_default': 100.,
          'f150': 90.
        }
      }

      lookup_conditional(source, 'my_param')
        => 100.

      lookup_conditional(source, 'my_param', tags=['f090'])
        => 100.

      lookup_conditional(source, 'my_param', tags=['f150'])
        => 90.

      lookup_conditional(source, 'my_other_param')
        KeyError!

      lookup_conditional(source, 'my_other_param', default=0)
        => 0

      # Note _default takes precedence over default argument.
      lookup_conditional(source, 'my_param', default=0)
        => 100.

      # Nested example:
      source = {
        'fit_params': {
          '_default': {
            'a': 12,
            'b': 100,
            '_stop_here': None,  # don't descend any further.
          },
          'f150': {
            'SAT': {
              'a': 1000,
              'b': 1200,
              '_stop_here': None,
            },
            'LAT': {
              'a': 1,
              'b': 2,
              '_stop_here': None,
            },
          },
        },
      }

      lookup_conditional(source, 'fit_params', tags=['f150', 'LAT'])
        => {'a': 1, 'b': 2}

      lookup_conditional(source, 'fit_params', tags=['LAT'])
        => {'a': 12, 'b': 100}

      lookup_conditional(source, 'fit_params', tags=['f150'])
        => {'a': 12, 'b': 100}

    """
    if tags is None:
        tags = []
    if key is not None:
        # On entry, key is not None.
        result = default
        if key in source:
            result = lookup_conditional(source[key], None, tags=tags, default=default)
        if inspect.isclass(result) and issubclass(result, Exception):
            raise result(f"Failed to find key '{key}' in {source}")
        return result
    else:
        # This block is entered on recursion.
        if not isinstance(source, dict):
            return source
        if '_stop_here' in source:
            return _filter_dict(source)
        # Update default?
        if '_default' in source:
            default = _filter_dict(source['_default'])
        # Find a tag.
        for t in tags:
            if t in source:
                return lookup_conditional(source[t], None, tags=tags, default=default)
        return default

class _ReltimeFormatter(logging.Formatter):
    def __init__(self, *args, t0=None, **kw):
        super().__init__(*args, **kw)
        if t0 is None:
            t0 = time.time()
        self.start_time = t0

    def formatTime(self, record, datefmt=None):
        if datefmt is None:
            datefmt = '%8.3f'
        return datefmt % (record.created - self.start_time)

def init_logger(name, announce='', verbosity=2, logger=None):
    """Configure and return a logger for site_pipeline elements.  It is
    disconnected from general sotodlib (propagate=False) and displays
    relative instead of absolute timestamps.

    """
    if logger is None:
        logger = logging.getLogger(name)

    if verbosity == 0:
        level = logging.ERROR
    elif verbosity == 1:
        level = logging.WARNING
    elif verbosity == 2:
        level = logging.INFO
    elif verbosity == 3:
        level = logging.DEBUG

    # add handler only if it doesn't exist
    if len(logger.handlers) == 0:
        ch = logging.StreamHandler(sys.stdout)
        formatter = _ReltimeFormatter('%(asctime)s: %(message)s (%(levelname)s)')

        ch.setLevel(level)
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        i, r = formatter.start_time // 1, formatter.start_time % 1
        text = (time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(i))
              + (',%03d' % (r*1000)))
        logger.info(f'{announce}Log timestamps are relative to {text}')
    else:
        for handler in logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                handler.setLevel(level)
                break

    logger.propagate = False
    logger.setLevel(logging.DEBUG)

    return logger

def main_launcher(main_func, parser_func, args=None):
    """Launch an element's main entry point function, after generating
    a parser and executing it on the command line arguments (or args
    if it is passed in).

    Args:
      main_func: the main entry point for a pipeline element.
      parser_func: the argument parser generation function for a pipeline
        element.
      args (list of str): arguments to parse (default is None, which
        will lead to sys.argv[1:]).

    Returns:
      Whatever main_func returns.

    """
    if args is None:
        args = sys.argv[1:]
    return main_func(**vars(parser_func().parse_args(args=args)))

def get_obslist(context, query=None, obs_id=None, min_ctime=None, max_ctime=None, 
                update_delay=None, tags=None, planet_obs=False):
    """Query the obs database with a given query.
    
    Parameters
    ----------
    context : core.Context
        The context to use for the obsdb.
    query : str, optional
        A query string for the obsdb.
    obs_id : str, optional
        The specific obsid to retrieve.
    min_ctime : int, optional
        The minimum ctime of obs to retrieve.
    max_ctime : int, optional
        The maximum ctime of obs to retrieve.
    update_delay : int, optional
        The number of days to subtract from the current time to set the minimum ctime.
    tags : list of str, optional
        A list of tags to use for the query.
    planet_obs : bool, optional
        If True, format query and tags for planet obs.
    
    Returns
    -------
    obs_list : list
        The list of obs found from the query.
    """

    try:
        with open(query, "r") as fname:
            return [context.obsdb.get(line.split()[0]) for line in fname]
    except FileNotFoundError:
        if (min_ctime is None) and (update_delay is not None):
            # If min_ctime is provided it will use that..
            # Otherwise it will use update_delay to set min_ctime.
            min_ctime = int(time.time()) - update_delay*86400

        if obs_id is not None:
            tot_query = f"obs_id=='{obs_id}'"
        else:
            tot_query = "and "
            if min_ctime is not None:
                tot_query += f"timestamp>={min_ctime} and "
            if max_ctime is not None:
                tot_query += f"timestamp<={max_ctime} and "
            if query is not None:
                tot_query += query + " and "
            tot_query = tot_query[4:-4]
            if tot_query=="":
                tot_query="1"

        if not(tags is None):
            for i, tag in enumerate(tags):
                tags[i] = tag.lower()
                if '=' not in tag:
                    tags[i] += '=1'

        if planet_obs:
            obs_list = []
            for tag in tags:
                obs_list.extend(context.obsdb.query(tot_query, tags=[tag]))
        else:
            obs_list = context.obsdb.query(tot_query, tags=tags)
        
        return obs_list

def send_alert(webhook, alertname='', tag='', error='', timestamp=None):
    """Send an alert to the given webhook.
    
    Parameters
    ----------
    webhook : str
        Webhook URL to send alert.
    alertname : str, optional
        Name of alert.
    tag : str, optional
        Tag to differentiate types of alerts.
    error : str, optional
        Error message to send as the main body of the alert.
    timestamp : str, optional
        The timestamp to attach to alert.
    """
    if not webhook:
        return "No webhook provided"
        
    if isinstance(timestamp, type(None)):
        timestamp = round(dt.datetime.now(tz=dt.timezone.utc).timestamp())
    elif isinstance(timestamp, str):
        ts = dt.datetime.fromisoformat(timestamp)
        if ts.tzinfo is None:
            timestamp = round(ts.replace(tzinfo=dt.timezone.utc).timestamp())
        else:
            timestamp = round(ts.timestamp())
    elif isinstance(timestamp, (int, float)):
        timestamp = round(timestamp)
    elif isinstance(timestamp, dt.datetime):
        if timestamp.tzinfo is None:
            timestamp = round(timestamp.replace(tzinfo=dt.timezone.utc).timestamp())
        else:
            timestamp = round(timestamp.timestamp())
    else:
        return f"Could not convert timestamp type {type(timestamp)}"
    
    try:
        # Directly to Slack channel
        if 'slack.com' in webhook:
            from slack_sdk.webhook import WebhookClient

            webhook = WebhookClient(webhook)
            response = webhook.send(
                text="Pipeline alert",
                attachments=[
                    {
                        "mrkdwn_in": ["text"],
                        "color": "danger",
                        "title": f"[FIRING] {tag} ({alertname})",
                        "text": f"error: {error}\ntimestamp: {timestamp}",
                        "footer": "Pipeline",
                        "ts": f"{time.time()}"
                    }
                ]
            )
            if response.status_code != 200:
                raise Exception("Issue with Slack webhook.")

        # Custom webhook (Campana)
        else:
            data = {
                    'status': 'firing',
                    'alerts': [
                        {
                            'status': 'firing',
                            'labels': {'alertname': alertname, 'tag': tag},
                            'annotations': {'error': error, 'groups': 'pipeline', 'timestamp': timestamp},
                        }
                    ],
            }
            response = requests.post(webhook, json=data)
            if response.status_code != requests.codes.ok:
                raise Exception("Issue with custom webhook.")
                
        return "Alert sent"
    except Exception as e:
        return f"Failed to send alert: {e}"
    


"""
Script for running updates on (or creating) a g3tsmurf database. This setup
is specifically designed to work when the data is dynamically coming in. Meaning is 
is designed to work from something like a cronjob. 
"""
import os
import yaml
import datetime as dt
import numpy as np
import argparse
import logging
from sqlalchemy import not_, or_, and_, desc

from sotodlib.io.g3thk_db import G3tHk, HKFiles, logger


def main(config=None, from_scratch=False, verbosity=2):

    show_pb = True if verbosity > 1 else False

    if verbosity == 0:
        logger.setLevel(logging.ERROR)
    elif verbosity == 1:
        logger.setLevel(logging.WARNING)
    elif verbosity == 2:
        logger.setLevel(logging.INFO)
    elif verbosity == 3:
        logger.setLevel(logging.DEBUG)

    HK = G3tHk.from_configs(config)

    if from_scratch or HK.session.query(HKFiles).count()==0:
        logger.info("Building Database from Scratch, May take awhile")
        min_time = int(1.6e9)
    else:
        ## start at the last file in the database
        last_file = HK.session.query(HKFiles)
        last_file = last_file.order_by(desc(HKFiles.global_start_time)).first()
        min_time = last_file.global_start_time - 10

    HK.add_hkfiles(min_ctime=min_time, show_pb=show_pb)

def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument('config', help="g3tsmurf db configuration file")
    parser.add_argument('--from-scratch', help="Builds or updates database from scratch",
                        action="store_true")
    parser.add_argument("--verbosity", help="increase output verbosity. 0:Error, 1:Warning, 2:Info(default), 3:Debug",
                       default=2, type=int)
    return parser



if __name__ == '__main__':
    parser = get_parser(parser=None)
    args = parser.parse_args()
    main(**vars(args))



# Copyright (c) 2018-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import h5py
import os
import pickle

import traitlets

import numpy as np

from astropy import constants
from astropy import units as u

import ephem

import healpy as hp

from scipy.constants import au as AU
from scipy.interpolate import RectBivariateSpline, splrep, splev

import toast.rng
from toast.timing import function_timer
from toast import qarray as qa
from toast.data import Data
from toast.traits import trait_docs, Float, Int, Unicode, Instance, Bool
from toast.ops.operator import Operator
from toast.utils import Logger, unit_conversion
from toast.observation import default_values as defaults


XAXIS, YAXIS, ZAXIS = np.eye(3)


@trait_docs
class SimHWPSS(Operator):
    """Simulate HWP synchronous signal.

    NOTE:  The HWPSS template is interpolated (with a 5th order spline interpolation).
    Interpolation errors produce spurious peaks in the frequency domain which are 7 to
    8 orders of magnitude below the amplitude of the original signal.

    """

    API = Int(0, help="Internal interface version for this operator")

    times = Unicode(
        defaults.times,
        help="Observation shared key for timestamps, only used for drift.",
    )

    hwp_angle = Unicode(
        defaults.hwp_angle, help="Observation shared key for HWP angle"
    )

    det_data = Unicode(
        defaults.det_data,
        help="Observation detdata key for simulated signal",
    )

    atmo_data = Unicode(
        None,
        allow_none=True,
        help="Observation detdata key for simulated atmosphere "
        "(modulates part of the HWPSS)",
    )

    stokes_weights = Instance(
        klass=Operator,
        allow_none=True,
        help="This must be an instance of a Stokes weights operator",
    )

    fname_hwpss = Unicode(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "data/hwpss_per_chi.pck",
        ),
        help="File containing measured or estimated HWPSS profiles",
    )

    drift_rate = Float(
        None,
        allow_none=True,
        help="If non-zero, the width of the Gaussian distribution to draw "
        "drift rate [1/hour] from.  All detectors will observe the same "
        "drift rate."
    )

    hwpss_random_drift = Bool(
        False,
        help="If True, the hwpss drift will be a random signal"
        "following a 1/f^alpha spectrum, and fully correlated between "
        "detectors."
    )

    hwpss_drift_alpha = Float(
        1.0,
        help="The power law exponent of the HWPSS random drift."
    )

    hwpss_drift_rms = Float(
        0.01,
        help="RMS of the relative HWPSS fluctuations."
    )

    hwpss_drift_coupling_center = Float(
        1.0,
        help="Mean coupling strength between the detectors and the HWPSS "
        "random drift mode."
    )

    hwpss_drift_coupling_width = Float(
        0.0,
        help="Width of the coupling strength distribution between the "
        "detectors and the HWPSS random drift mode."
    )

    realization = Int(0, help="Realization ID")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        for trait in "stokes_weights",:
            value = getattr(self, trait)
            if value is None:
                raise RuntimeError(
                    f"You must set `{trait}` before running SimHWPSS"
                )

        if not os.path.isfile(self.fname_hwpss):
            raise RuntimeError(f"{self.fname_hwpss} does not exist!")

        # theta is the incident angle, also known as the radial distance
        #     to the boresight.
        # chi is the HWP rotation angle

        with open(self.fname_hwpss, "rb") as fin:
            self.thetas, self.chis, self.all_stokes = pickle.load(fin)

        for obs in data.obs:
            dets = obs.select_local_detectors(detectors)
            obs.detdata.ensure(self.det_data, detectors=dets, create_units=u.K)
            det_units = obs.detdata[self.det_data].units
            det_scale = unit_conversion(u.K, det_units)

            focalplane = obs.telescope.focalplane
            fs = focalplane.sample_rate.to_value(u.Hz)

            if self.drift_rate is not None and self.drift_rate != 0:
                # Randomize the drift in a reproducible manner
                counter1 = obs.session.uid
                counter2 = self.realization
                key1 = 683584
                key2 = 476365
                x = toast.rng.random(
                    1,
                    sampler="gaussian",
                    key=(key1, key2),
                    counter=(counter1, counter2),
                )[0]
                drift_rate = self.drift_rate * x
                # Translate to actual drift
                t = obs.shared[self.times].data
                tmean = np.mean(t)  # Assumes data distribution by detector
                drift = drift_rate * (t - tmean) / 3600
            elif self.hwpss_random_drift and self.hwpss_drift_alpha is not None:
                # Generate the HWPSS random drift common mode
                counter1 = obs.session.uid
                counter2 = self.realization
                key1 = 683584
                key2 = 476365
                nsamp = obs.shared[self.times].data.size
                w = toast.rng.random(
                    nsamp,
                    sampler="gaussian",
                    key=(key1, key2),
                    counter=(counter1, counter2),
                ).array()
                freqs = np.fft.rfftfreq(nsamp, 1/fs)
                drift_psd = np.zeros_like(freqs)
                drift_psd[1:] = np.abs(1 / freqs[1:])**self.hwpss_drift_alpha
                drift = np.fft.irfft(np.fft.rfft(w) * np.sqrt(drift_psd))
                drift *= self.hwpss_drift_rms / np.std(drift)
            else:
                drift = 0

            # Get HWP angle
            chi = obs.shared[self.hwp_angle].data
            for det in dets:
                signal = obs.detdata[self.det_data][det]
                if self.atmo_data is None:
                    atmo = None
                else:
                    atmo = obs.detdata[self.atmo_data][det]
                band = focalplane[det]["band"]
                freq = {
                    "SAT_f030" : "027",
                    "SAT_f040" : "039",
                    "SAT_f090" : "093",
                    "SAT_f150" : "145",
                    "SAT_f230" : "225",
                    "SAT_f290" : "278",
                }[band]

                # Get incident angle

                det_quat = focalplane[det]["quat"]
                det_theta, det_phi, det_psi = qa.to_iso_angles(det_quat)

                # Compute Stokes weights

                iweights = 1
                qweights = np.cos(2 * det_psi)
                uweights = np.sin(2 * det_psi)

                # Convert Az/El quaternion of the detector into elevation

                obs_data = Data(comm=data.comm)
                obs_data._internal = data._internal
                obs_data.obs = [obs]
                self.stokes_weights.apply(obs_data, detectors=[det])
                obs_data.obs.clear()
                del obs_data

                azel_quat = obs.detdata[
                    self.stokes_weights.detector_pointing.quats
                ][det]
                theta, phi, _ = qa.to_iso_angles(azel_quat)
                el = np.pi / 2 - theta

                # Get polarization weights

                weights = obs.detdata[self.stokes_weights.weights][det]
                iweights, qweights, uweights = weights.T

                # Interpolate HWPSS to incident angle equal to the
                # radial distance from the focalplane (HWP) center

                theta_deg = np.degrees(det_theta)
                itheta_high = np.searchsorted(self.thetas, theta_deg)
                itheta_low = itheta_high - 1

                theta_low = self.thetas[itheta_low]
                theta_high = self.thetas[itheta_high]
                r = (theta_deg - theta_low) / (theta_high - theta_low)

                # HWPSS not from atmosphere

                transmission_wo_atmo = (
                    (1 - r) * self.all_stokes[freq]["transmission_wo_atmo"][itheta_low]
                    + r * self.all_stokes[freq]["transmission_wo_atmo"][itheta_high]
                )
                reflection_wo_atmo = (
                    (1 - r) * self.all_stokes[freq]["reflection_wo_atmo"][itheta_low]
                    + r * self.all_stokes[freq]["reflection_wo_atmo"][itheta_high]
                )
                # Thermal emission from the HWP is not driven by the atmosphere
                emission = (
                    (1 - r) * self.all_stokes[freq]["emission_wo_atmo"][itheta_low]
                    + r * self.all_stokes[freq]["emission_wo_atmo"][itheta_high]
                )

                # HWPSS from atmosphere

                transmission_atmo = (
                    (1 - r) * self.all_stokes[freq]["transmission_atmo"][itheta_low]
                    + r * self.all_stokes[freq]["transmission_atmo"][itheta_high]
                )
                reflection_atmo = (
                    (1 - r) * self.all_stokes[freq]["reflection_atmo"][itheta_low]
                    + r * self.all_stokes[freq]["reflection_atmo"][itheta_high]
                )

                if atmo is None:
                    transmission = transmission_wo_atmo + transmission_atmo
                    reflection = reflection_wo_atmo + reflection_atmo
                else:
                    transmission = transmission_wo_atmo
                    reflection = reflection_wo_atmo

                # Scale HWPSS for observing elevation

                el_ref = np.radians(50)
                scale = np.sin(el_ref) / np.sin(el)

                # Observe HWPSS with the detector

                iquv = (transmission + reflection)
                iquv = iquv.T.copy()
                iquss = (
                    iweights * splev(chi, splrep(self.chis, iquv[0], k=5)) +
                    qweights * splev(chi, splrep(self.chis, iquv[1], k=5)) +
                    uweights * splev(chi, splrep(self.chis, iquv[2], k=5))
                ) * scale

                if atmo is not None:
                    # Atmospheric HWPSS is modulated by the relative
                    # atmospheric fluctuation
                    modulation = atmo / np.median(atmo)
                    iquv = (transmission_atmo + reflection_atmo)
                    iquv = iquv.T.copy()
                    # Replace the generic T offset with the simulated atmosphere
                    # offset
                    iquv[0] += np.median(atmo) / det_scale - np.mean(iquv[0])
                    iquss += (
                        iweights * splev(chi, splrep(self.chis, iquv[0], k=5)) +
                        qweights * splev(chi, splrep(self.chis, iquv[1], k=5)) +
                        uweights * splev(chi, splrep(self.chis, iquv[2], k=5))
                    ) * scale * modulation

                iquv = (emission).T.copy()
                iquss += (
                    iweights * splev(chi, splrep(self.chis, iquv[0], k=5)) +
                    qweights * splev(chi, splrep(self.chis, iquv[1], k=5)) +
                    uweights * splev(chi, splrep(self.chis, iquv[2], k=5))
                )

                if self.hwpss_random_drift:
                    # Apply detector couplings to HWPSS random drift common mode
                    key1 = obs.telescope.uid
                    key2 = obs.session.uid
                    counter1 = self.realization
                    counter2 = focalplane[det]["uid"]
                    gaussian = toast.rng.random(
                        1,
                        sampler="gaussian",
                        key=(key1, key2),
                        counter=(counter1, counter2),
                    )[0]
                    coupling = (
                        self.hwpss_drift_coupling_center
                        + gaussian * self.hwpss_drift_coupling_width
                    )
                else:
                    coupling = 1.0

                # Co-add with the cached signal

                if atmo is not None:
                    # Avoid double-counting the atmosphere.
                    # HWPSS has a copy of it
                    signal -= atmo
                signal += det_scale * iquss * (1 + drift * coupling)

        return

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = self.stokes_weights.requires()
        req["shared"].append(self.hwp_angle)
        req["detdata"].append(self.weights)
        return req

    def _provides(self):
        prov = {
            "meta": list(),
            "shared": list(),
            "detdata": [
                self.det_data,
            ],
        }
        return prov

    def _accelerators(self):
        return list()



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Timestream processing filters.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_raw_statistics(operators):
    """Add commandline args and operators for raw timestream statistics.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.Statistics(name="raw_statistics", enabled=False))


@workflow_timer
def raw_statistics(job, otherargs, runargs, data):
    """Compute timestream statistics on the raw data.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.raw_statistics.enabled:
        job_ops.raw_statistics.output_dir = otherargs.out_dir
        job_ops.raw_statistics.apply(data)


def setup_filtered_statistics(operators):
    """Add commandline args and operators for filtered timestream statistics.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.Statistics(name="filtered_statistics", enabled=False))


@workflow_timer
def filtered_statistics(job, otherargs, runargs, data):
    """Compute timestream statistics on the filtered data.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.filtered_statistics.enabled:
        job_ops.filtered_statistics.output_dir = otherargs.out_dir
        job_ops.filtered_statistics.apply(data)


def setup_hn_map(operators):
    """Add commandline args and operators for H_n map.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.Hn(name="h_n", enabled=False))


@workflow_timer
def hn_map(job, otherargs, runargs, data):
    """Compute the H_n map.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.h_n.enabled:
        job_ops.h_n.pixel_pointing = job.pixels_final
        job_ops.h_n.pixel_dist = job_ops.binner_final.pixel_dist
        job_ops.h_n.output_dir = otherargs.out_dir
        job_ops.h_n.save_pointing = otherargs.full_pointing
        job_ops.h_n.apply(data)


def setup_cadence_map(operators):
    """Add commandline args and operators for the cadence map.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.CadenceMap(name="cadence_map", enabled=False))


@workflow_timer
def cadence_map(job, otherargs, runargs, data):
    """Compute the cadence map.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.cadence_map.enabled:
        job_ops.cadence_map.pixel_pointing = job.pixels_final
        job_ops.cadence_map.pixel_dist = job_ops.binner_final.pixel_dist
        job_ops.cadence_map.output_dir = otherargs.out_dir
        job_ops.cadence_map.save_pointing = otherargs.full_pointing
        job_ops.cadence_map.apply(data)


def setup_crosslinking_map(operators):
    """Add commandline args and operators for the crosslinking map.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.CrossLinking(name="crosslinking", enabled=False))


@workflow_timer
def crosslinking_map(job, otherargs, runargs, data):
    """Compute the crosslinking map.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.crosslinking.enabled:
        job_ops.crosslinking.pixel_pointing = job.pixels_final
        job_ops.crosslinking.pixel_dist = job_ops.binner_final.pixel_dist
        job_ops.crosslinking.output_dir = otherargs.out_dir
        job_ops.crosslinking.save_pointing = otherargs.full_pointing
        job_ops.crosslinking.apply(data)



import os
import logging
import time
import sys
import copy
import yaml
import numpy as np
import h5py
import traceback
import inspect
from pathlib import Path
import re
from sotodlib.hwp import hwp_angle_model
from sotodlib.coords import demod as demod_mm
from sotodlib.tod_ops import t2pleakage
from sotodlib.core.flagman import has_any_cuts

from .. import core

from . import _Preprocess, Pipeline, processes

class ArchivePolicy:
    """Storage policy assistance.  Helps to determine the HDF5
    filename and dataset name for a result.

    Make me better!

    """
    @staticmethod
    def from_params(params):
        if params['type'] == 'simple':
            return ArchivePolicy(**params)
        if params['type'] == 'directory':
            return DirectoryArchivePolicy(**params)
        raise ValueError('No handler for "type"="%s"' % params['type'])

    def __init__(self, **kwargs):
        self.filename = kwargs['filename']

    def get_dest(self, product_id):
        """Returns (hdf_filename, dataset_addr).

        """
        return self.filename, product_id


class DirectoryArchivePolicy:
    """Storage policy for stuff organized directly on the filesystem.

    """
    def __init__(self, **kwargs):
        self.root_dir = kwargs['root_dir']
        self.pattern = kwargs['pattern']

    def get_dest(self, **kw):
        """Returns full path to destination directory.

        """
        return os.path.join(self.root_dir, self.pattern.format(**kw))


class _ReltimeFormatter(logging.Formatter):
    def __init__(self, *args, t0=None, **kw):
        super().__init__(*args, **kw)
        if t0 is None:
            t0 = time.time()
        self.start_time = t0

    def formatTime(self, record, datefmt=None):
        if datefmt is None:
            datefmt = '%8.3f'
        return datefmt % (record.created - self.start_time)


def init_logger(name, announce='', verbosity=2):
    """Configure and return a logger for site_pipeline elements.  It is
    disconnected from general sotodlib (propagate=False) and displays
    relative instead of absolute timestamps.

    Arguments
    ----------
    name : str
        The name of the logger
    announce : str
        Initial message to be displayed after logger is instantiated.
    verbosity : int
        Level of logger output
        0: Error
        1: Warning
        2: Info
        3: Debug

    Returns
    -------
    logger : PythonLogger
        The initialized logger object
    """

    logger = logging.getLogger(name)

    if verbosity == 0:
        level = logging.ERROR
    elif verbosity == 1:
        level = logging.WARNING
    elif verbosity == 2:
        level = logging.INFO
    elif verbosity == 3:
        level = logging.DEBUG

    # add handler only if it doesn't exist
    if len(logger.handlers) == 0:
        ch = logging.StreamHandler(sys.stdout)
        formatter = _ReltimeFormatter('%(asctime)s: %(message)s (%(levelname)s)')

        ch.setLevel(level)
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        i, r = formatter.start_time // 1, formatter.start_time % 1
        text = (time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(i))
              + (',%03d' % (r*1000)))
        logger.info(f'{announce}Log timestamps are relative to {text}')
    else:
        for handler in logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                handler.setLevel(level)
                break

    logger.propagate = False
    logger.setLevel(logging.DEBUG)

    return logger


def get_preprocess_context(configs, context=None):
    """Load the provided config file and context file. To be used in
    ``preprocess_*.py`` site pipeline scripts. If the provided context
    file does not have a metadata entry for preprocess then one will
    be added based on the definition in the config file.

    Arguments
    ----------
    configs : str or dict
        The configuration file or dictionary.
    context : str or core.Context, optional
        The context to use. If None, it is created from the configuration file.

    Returns
    -------
    configs : dict
        The configuration dictionary.
    context : core.Context
        The context file.
    """

    if type(configs) == str:
        configs = yaml.safe_load(open(configs, "r"))

    if context is None:
        context = core.Context(configs["context_file"])

    if type(context) == str:
        context = core.Context(context)

    # if context doesn't have the preprocess archive it in add it
    # allows us to use same context before and after calculations
    found=False
    if context.get("metadata") is None:
        context["metadata"] = []

    for key in context.get("metadata"):
        if key.get("name") == "preprocess" or key.get("label") == "preprocess":
            found=True
            break
    if not found:
        context["metadata"].append(
            {
                "db" : configs["archive"]["index"],
                "name" : "preprocess"
            }
        )
    return configs, context


def get_groups(obs_id, configs, context):
    """Get subobs group method and groups. To be used in
    ``preprocess_*.py`` site pipeline scripts.

    Arguments
    ----------
    obs_id : str
        The obsid.
    configs : dict
        The configuration dictionary.
    context : core.Context
        The Context file to use.

    Returns
    -------
    group_by : list of str
        The list of keys used to group the detectors.
    groups : list of list of int
        The list of groups of detectors.
    """

    try:
        group_by = np.atleast_1d(configs['subobs'].get('use', 'detset'))
        for i, gb in enumerate(group_by):
            if gb.startswith('dets:'):
                group_by[i] = gb.split(':',1)[1]

            if (gb == 'detset') and (len(group_by) == 1):
                groups = context.obsfiledb.get_detsets(obs_id)
                return group_by, [[g] for g in groups], None

        det_info = context.get_det_info(obs_id)
        rs = det_info.subset(keys=group_by).distinct()
        groups = [[b for a,b in r.items()] for r in rs]
        return group_by, groups, None
    except Exception as e:
        error = f'Failed get groups for: {obs_id}'
        errmsg = f'{type(e)}: {e}'
        tb = ''.join(traceback.format_tb(e.__traceback__))
        return [], [], [error, errmsg, tb]


def get_preprocess_db(configs, group_by, logger=None):
    """Get or create a ManifestDb found for a given
    config.

    Arguments
    ----------
    configs : dict
        The configuration dictionary.
    group_by : list of str
        The list of keys used to group the detectors.
    logger : PythonLogger
        Optional. Logger object.  If None, a new logger
        is created.

    Returns
    -------
    db : ManifestDb
        ManifestDb object
    """

    if logger is None:
        logger = init_logger("preprocess_db")

    if os.path.exists(configs['archive']['index']):
        logger.info(f"Mapping {configs['archive']['index']} for the "
                    "archive index.")
        db = core.metadata.ManifestDb(configs['archive']['index'])
    else:
        logger.info(f"Creating {configs['archive']['index']} for the "
                     "archive index.")
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('obs:obs_id')
        for gb in group_by:
            scheme.add_exact_match('dets:' + gb)
        scheme.add_data_field('dataset')
        db = core.metadata.ManifestDb(
            configs['archive']['index'],
            scheme=scheme
        )
    return db


def swap_archive(config, fpath):
    """Update the configuration archive policy filename,
    create an output archive directory if it doesn't exist,
    and return a copy of the config.

    Arguments
    ----------
    configs : dict
        The configuration dictionary.
    fpath : str
        The archive policy filename to write to.

    Returns
    -------
    tc : dict
        Copy of the configuration file with an updated archive policy filename
    """

    tc = copy.deepcopy(config)
    tc['archive']['policy']['filename'] = os.path.join(os.path.dirname(tc['archive']['policy']['filename']), fpath)
    dname = os.path.dirname(tc['archive']['policy']['filename'])
    if not(os.path.exists(dname)):
        os.makedirs(dname)
    return tc


def load_preprocess_det_select(obs_id, configs, context=None,
                               dets=None, meta=None, logger=None):
    """Loads the metadata information for the Observation and runs through any
    data selection specified by the Preprocessing Pipeline.

    Arguments
    ----------
    obs_id: multiple
        Passed to `context.get_obs` to load AxisManager, see Notes for
        `context.get_obs`
    configs: string or dictionary
        Config file or loaded config directory
    context: core.Context
        The Context file to use.
    dets: dict
        Dets to restrict on from info in det_info. See context.get_meta.
    meta: AxisManager
        Contains supporting metadata to use for loading.
        Can be pre-restricted in any way. See context.get_meta.
    logger: PythonLogger
        Optional. Logger object.  If None, a new logger
        is created.
    """

    if logger is None:
        logger = init_logger("preprocess")

    configs, context = get_preprocess_context(configs, context)
    pipe = Pipeline(configs["process_pipe"], logger=logger)

    if meta is None:
        meta = context.get_meta(obs_id, dets=dets)
    logger.info("Restricting detectors on all processes")
    keep_all = np.ones(meta.dets.count,dtype=bool)
    for process in pipe[:]:
        keep = process.select(meta, in_place=False)
        if isinstance(keep, np.ndarray):
            keep_all &= keep
    return meta.dets.vals[keep_all]


def load_and_preprocess(obs_id, configs, context=None, dets=None, meta=None,
                        no_signal=None, logger=None):
    """Loads the saved information from the preprocessing pipeline and runs
    the processing section of the pipeline.

    Assumes preprocess_tod has already been run on the requested observation.

    Arguments
    ----------
    obs_id: multiple
        Passed to `context.get_obs` to load AxisManager, see Notes for
        `context.get_obs`
    configs: string or dictionary
        Config file or loaded config directory
    context: core.Context
        Optional. The Context file to use.
    dets: dict
        Dets to restrict on from info in det_info. See context.get_meta.
    meta: AxisManager
        Contains supporting metadata to use for loading.
        Can be pre-restricted in any way. See context.get_meta.
    no_signal: bool
        If True, signal will be set to None.
        This is a way to get the axes and pointing info without
        the (large) TOD blob.  Not all loaders may support this.
    logger: PythonLogger
        Optional. Logger object.  If None, a new logger
        is created.
    """

    if logger is None:
        logger = init_logger("preprocess")

    configs, context = get_preprocess_context(configs, context)
    meta = context.get_meta(obs_id, dets=dets, meta=meta)
    if (
        'valid_data' in meta.preprocess and
        isinstance(meta.preprocess.valid_data, core.AxisManager)
       ):
        keep = has_any_cuts(meta.preprocess.valid_data.valid_data)
        meta.restrict("dets", keep)
    else:
        det_vals = load_preprocess_det_select(obs_id, configs=configs, context=context,
                                              dets=dets, meta=meta, logger=logger)
        meta.restrict("dets", [d for d in meta.dets.vals if d in det_vals])

    if meta.dets.count == 0:
        logger.info(f"No detectors left after cuts in obs {obs_id}")
        return None
    else:
        pipe = Pipeline(configs["process_pipe"], logger=logger)
        aman = context.get_obs(meta, no_signal=no_signal)
        pipe.run(aman, aman.preprocess, select=False)
        return aman


def multilayer_load_and_preprocess(obs_id, configs_init, configs_proc,
                                   dets=None, meta=None, no_signal=None,
                                   logger=None, init_only=False):
    """Loads the saved information from the preprocessing pipeline from a
    reference and a dependent database and runs the processing section of
    the pipeline for each.

    Assumes preprocess_tod and multilayer_preprocess_tod have already been run
    on the requested observation.

    Arguments
    ----------
    obs_id: multiple
        Passed to `context.get_obs` to load AxisManager, see Notes for
        `context.get_obs`
    configs_init: string or dictionary
        Config file or loaded config directory
    configs_proc: string or dictionary
        Second config file or loaded config dictionary to load
        dependent databases generated using multilayer_preprocess_tod.py.
    dets: dict
        Dets to restrict on from info in det_info. See context.get_meta.
    meta: AxisManager
        Contains supporting metadata to use for loading.
        Can be pre-restricted in any way. See context.get_meta.
    no_signal: bool
        If True, signal will be set to None.
        This is a way to get the axes and pointing info without
        the (large) TOD blob.  Not all loaders may support this.
    logger: PythonLogger
        Optional. Logger object or None will generate a new one.
    init_only: bool
        Optional. If True, do not run the dependent pipeline.
    """

    if logger is None:
        logger = init_logger("preprocess")

    configs_init, context_init = get_preprocess_context(configs_init)
    meta_init = context_init.get_meta(obs_id, dets=dets, meta=meta)

    configs_proc, context_proc = get_preprocess_context(configs_proc)
    meta_proc = context_proc.get_meta(obs_id, dets=dets, meta=meta)

    group_by_init, groups_init, error_init = get_groups(obs_id, configs_init, context_init)
    group_by_proc, groups_proc, error_proc = get_groups(obs_id, configs_proc, context_proc)

    if error_init is not None:
        raise ValueError(f"{error_init[0]}\n{error_init[1]}\n{error_init[2]}")

    if error_proc is not None:
        raise ValueError(f"{error_proc[0]}\n{error_proc[1]}\n{error_proc[2]}")

    if (group_by_init != group_by_proc).any():
        raise ValueError('init and proc groups do not match')

    if meta_init.dets.count == 0 or meta_proc.dets.count == 0:
        logger.info(f"No detectors in obs {obs_id}")
        return None
    else:
        pipe_init = Pipeline(configs_init["process_pipe"], logger=logger)
        aman_cfgs_ref = get_pcfg_check_aman(pipe_init)

        if check_cfg_match(aman_cfgs_ref, meta_proc.preprocess['pcfg_ref'],
                           logger=logger):
            pipe_proc = Pipeline(configs_proc["process_pipe"], logger=logger)

            logger.info("Restricting detectors on all proc pipeline processes")
            if (
                'valid_data' in meta_proc.preprocess and
                isinstance(meta_proc.preprocess.valid_data, core.AxisManager)
               ):
                keep_all = has_any_cuts(meta_proc.preprocess.valid_data.valid_data)
            else:
                keep_all = np.ones(meta_proc.dets.count, dtype=bool)
                for process in pipe_proc[:]:
                    keep = process.select(meta_proc, in_place=False)
                    if isinstance(keep, np.ndarray):
                        keep_all &= keep
            meta_proc.restrict("dets", meta_proc.dets.vals[keep_all])
            meta_init.restrict('dets', meta_proc.dets.vals)

            aman = context_init.get_obs(meta_init, no_signal=no_signal)
            logger.info("Running initial pipeline")
            pipe_init.run(aman, aman.preprocess, select=False)
            if init_only:
                return aman

            logger.info("Running dependent pipeline")
            proc_aman = context_proc.get_meta(obs_id, meta=aman)

            if 'valid_data' in aman.preprocess:
                aman.preprocess.move('valid_data', None)
            aman.preprocess.merge(proc_aman.preprocess)
            pipe_proc.run(aman, aman.preprocess, select=False)

            return aman
        else:
            raise ValueError('Dependency check between configs failed.')


def multilayer_load_and_preprocess_sim(obs_id, configs_init, configs_proc,
                                       sim_map, meta=None,
                                       logger=None, init_only=False,
                                       t2ptemplate_aman=None):
    """Loads the saved information from the preprocessing pipeline from a
    reference and a dependent database, loads the signal from a (simulated)
    map into the AxisManager and runs the processing section of the pipeline
    for both databases.

    Assumes preprocess_tod and multilayer_preprocess_tod have already been run
    on the requested observation.

    Arguments
    ----------
    obs_id: multiple
        Passed to `context.get_obs` to load AxisManager, see Notes for
        `context.get_obs`
    configs_init: string or dictionary
        Config file or loaded config directory
    configs_proc: string or dictionary
        Second config file or loaded config dictionary to load
        dependent databases generated using multilayer_preprocess_tod.py.
    sim_map: numpy.ndmap or enmap.ndmap
        Input simulated map to be observed
    meta: AxisManager
        Contains supporting metadata to use for loading.
        Can be pre-restricted in any way. See context.get_meta.
    no_signal: bool
        If True, signal will be set to None.
        This is a way to get the axes and pointing info without
        the (large) TOD blob.  Not all loaders may support this.
    logger: PythonLogger
        Optional. Logger object or None will generate a new one.
    init_only: bool
        Optional. Whether or not to run the dependent pipeline.
    t2ptemplate_aman: AxisManager
        Optional. AxisManager to use as a template for t2p leakage
        deprojection.
    """
    if logger is None:
        logger = init_logger("preprocess")

    configs_init, context_init = get_preprocess_context(configs_init)
    meta_init = context_init.get_meta(obs_id, meta=meta)

    configs_proc, context_proc = get_preprocess_context(configs_proc)
    meta_proc = context_proc.get_meta(obs_id, meta=meta)

    group_by_init, groups_init, error_init = get_groups(obs_id, configs_init, context_init)
    group_by_proc, groups_proc, error_proc = get_groups(obs_id, configs_proc, context_proc)

    if error_init is not None:
        raise ValueError(f"{error_init[0]}\n{error_init[1]}\n{error_init[2]}")

    if error_proc is not None:
        raise ValueError(f"{error_proc[0]}\n{error_proc[1]}\n{error_proc[2]}")

    if (group_by_init != group_by_proc).any():
        raise ValueError('init and proc groups do not match')

    if meta_init.dets.count == 0 or meta_proc.dets.count == 0:
        logger.info(f"No detectors in obs {obs_id}")
        return None
    else:
        pipe_init = Pipeline(configs_init["process_pipe"], logger=logger)
        aman_cfgs_ref = get_pcfg_check_aman(pipe_init)

        if check_cfg_match(aman_cfgs_ref, meta_proc.preprocess['pcfg_ref'],
                           logger=logger):
            pipe_proc = Pipeline(configs_proc["process_pipe"], logger=logger)

            logger.info("Restricting detectors on all proc pipeline processes")
            keep_all = np.ones(meta_proc.dets.count, dtype=bool)
            for process in pipe_proc[:]:
                keep = process.select(meta_proc, in_place=False)
                if isinstance(keep, np.ndarray):
                    keep_all &= keep
            meta_proc.restrict("dets", meta_proc.dets.vals[keep_all])
            meta_init.restrict('dets', meta_proc.dets.vals)
            aman = context_init.get_obs(meta_proc, no_signal=True)
            aman = hwp_angle_model.apply_hwp_angle_model(aman)
            aman.move("signal", None)

            logger.info("Reading in simulated map")
            demod_mm.from_map(aman, sim_map, wrap=True, modulated=True)

            logger.info("Running initial pipeline")
            pipe_init.run(aman, aman.preprocess, sim=True)

            if init_only:
                return aman
            
            if t2ptemplate_aman is not None:
                # Replace Q,U with simulated timestreams
                t2ptemplate_aman.wrap("demodQ", aman.demodQ, [(0, 'dets'), (1, 'samps')], overwrite=True)
                t2ptemplate_aman.wrap("demodU", aman.demodU, [(0, 'dets'), (1, 'samps')], overwrite=True)

                t2p_aman = t2pleakage.get_t2p_coeffs(
                    t2ptemplate_aman,
                    merge_stats=False
                )
                t2pleakage.subtract_t2p(
                    aman,
                    t2p_aman,
                    T_signal=t2ptemplate_aman.dsT
                )

            logger.info("Running dependent pipeline")
            proc_aman = context_proc.get_meta(obs_id, meta=aman)
            if 'valid_data' in aman.preprocess:
                aman.preprocess.move('valid_data', None)
            aman.preprocess.merge(proc_aman.preprocess)
            pipe_proc.run(aman, aman.preprocess, sim=True)

            return aman
        else:
            raise ValueError('Dependency check between configs failed.')


def find_db(obs_id, configs, dets, context=None, logger=None):
    """This function checks if the manifest db from
    a config file exists and searches if it contains
    an entry for the provided Obs id and set of detectors.

    Arguments
    ----------
    obs_id: str
        Obs id to process or load
    configs: fpath or dict
        Filepath or dictionary containing the preprocess configuration file.
    dets: dict
        Dictionary specifying which detectors/wafers to load see ``Context.obsdb.get_obs``.
    context: core.Context
        Optional. Context object used for data loading/querying.
    logger: PythonLogger
        Optional. Logger object or None will generate a new one.

    Returns
    -------
    dbexist : bool
        True if db exists and entry for input detectors is found.
    """

    if logger is None:
        logger = init_logger("preprocess")

    if type(configs) == str:
        configs = yaml.safe_load(open(configs, "r"))
    if context is None:
        context = core.Context(configs["context_file"])
    group_by, _, _ = get_groups(obs_id, configs, context)
    cur_groups = [list(np.fromiter(dets.values(), dtype='<U32'))]
    dbexist = True
    if os.path.exists(configs['archive']['index']):
        db = core.metadata.ManifestDb(configs['archive']['index'])
        dbix = {'obs:obs_id':obs_id}
        for gb, g in zip(group_by, cur_groups[0]):
            dbix[f'dets:{gb}'] = g
        logger.info(f'find_db found: {dbix}')
        if len(db.inspect(dbix)) == 0:
            dbexist = False
    else:
        dbexist = False

    return dbexist


def cleanup_archive(configs, logger=None):
    """This function finds the final preprocess archive file and deletes any
    datasets that are not found in the preprocess database.  This helps avoid
    cases where the database writing was interrupted in a previous run.

    Arguments
    ----------
    configs: fpath or dict
        Filepath or dictionary containing the preprocess configuration file.
    logger: PythonLogger
        Optional. Logger object or None will generate a new one.
    """

    if type(configs) == str:
        configs = yaml.safe_load(open(configs, "r"))

    if logger is None:
        logger = init_logger("preprocess")

    if os.path.exists(configs['archive']['index']):
        db = core.metadata.ManifestDb(configs['archive']['index'])

        basename = os.path.splitext(os.path.basename(configs["archive"]["policy"]["filename"]))[0]

        # remove datasets from last archive file if they are not in db
        archive_files = list(
            Path(os.path.dirname(configs["archive"]["policy"]["filename"])).rglob(f"{basename}*.h5")
        )
        pattern = re.compile(r"\d+")
        archive_files = [p for p in archive_files if pattern.findall(p.stem)]

        if archive_files:
            latest_file = max([(int(pattern.findall(p.stem)[-1]), p)
                               for p in archive_files if pattern.findall(p.stem)],
                              key=lambda t: t[0])[1]

            db_datasets = [d['dataset'] for d in db.inspect()]
            with h5py.File(latest_file, "r+") as f:
                keys = list(f.keys())
                for key in keys:
                    if key not in db_datasets:
                        logger.debug(f"{key} not found in db. deleting it from {latest_file}.")
                        del f[key]

        db.conn.close()


def save_group(obs_id, configs, dets, context=None, subdir='temp'):
    """This function returns a dictionary containing the data destination filename
    and the values to populate the manifest db.

    Arguments
    ----------
    obs_id: str
        Obs id to process or load
    configs: fpath or dict
        Filepath or dictionary containing the preprocess configuration file.
    dets: dict
        Dictionary specifying which detectors/wafers to load see ``Context.obsdb.get_obs``.
    context: core.Context
        Optional. Context object used for data loading/querying.
    subdir: str
        Optional. Subdirectory to save the output files into.  If it does not exist, it is created.

    Returns
    -------
    outputs : dict
        Dictionary including output filename of data file and information for corresponding
        database entry.
    """

    if type(configs) == str:
        configs = yaml.safe_load(open(configs, "r"))
    if context is None:
        context = core.Context(configs["context_file"])

    cur_groups = [list(np.fromiter(dets.values(), dtype='<U32'))]
    group_by = np.atleast_1d(configs['subobs'].get('use', 'detset'))
    newpath = f'{subdir}/{obs_id}'
    for cg in cur_groups[0]:
        newpath += f'_{cg}'
    temp_config = swap_archive(configs, newpath+'.h5')
    policy = ArchivePolicy.from_params(temp_config['archive']['policy'])
    dest_file, dest_dataset = policy.get_dest(obs_id)
    for gb, g in zip(group_by, cur_groups[0]):
        if gb == 'detset':
            dest_dataset += "_" + g
        else:
            dest_dataset += "_" + gb + "_" + str(g)

    # Collect info for saving h5 file.
    outputs = {}
    outputs['temp_file'] = dest_file

    # Collect index info.
    db_data = {'obs:obs_id': obs_id,
                'dataset': dest_dataset}
    for gb, g in zip(group_by, cur_groups[0]):
        db_data['dets:'+gb] = g
    outputs['db_data'] = db_data

    return outputs


def save_group_and_cleanup(obs_id, configs, context=None, subdir='temp',
                           logger=None, remove=False):
    """This function checks if any temporary files exist from a preprocessing
     run and will either add them to the config policy file and create an entry
     in the manifest db by calling ``cleanup_mandb``.  If the file exists but
     cannot be opened or if remove is True, the file will be deleted. Remove
     is intended to be to allow for overwrite=True in ``preprocess_tod.py``
     and ``multilayer_preprocess_tod.py``.

    Arguments
    ----------
    obs_id: str
        Obs id to process or load
    configs: fpath or dict
        Filepath or dictionary containing the preprocess configuration file.
    context: core.Context
        Optional. Context object used for data loading/querying.
    subdir: str
        Optional. Subdirectory to save the output files into.  If it does not exist, it is created.
    logger: PythonLogger
        Optional. Logger object or None will generate a new one.
    remove: bool
        Optional. Default is False. Whether to remove a file if found.
        Used when ``overwrite`` is True in driving functions.
    """

    if logger is None:
        logger = init_logger("preprocess")

    if type(configs) == str:
        configs = yaml.safe_load(open(configs, "r"))

    if context is None:
        context = core.Context(configs["context_file"])

    group_by, groups, error = get_groups(obs_id, configs, context)

    if os.path.exists(configs['archive']['index']):
        db = core.metadata.ManifestDb(configs['archive']['index'])
        x = db.inspect({'obs:obs_id': obs_id})
    else:
        x = None

    all_groups = groups.copy()
    for g in all_groups:
        if 'wafer.bandpass' in group_by:
            if 'NC' in g:
                groups.remove(g)
                continue

    # get groups in db
    db_groups = []
    if x is not None and len(x) > 0:
        [db_groups.append([a[f'dets:{gb}'] for gb in group_by]) for a in x]

    for g in groups:
        dets = {gb:gg for gb, gg in zip(group_by, g)}
        outputs_grp = save_group(obs_id, configs, dets, context, subdir)

        if os.path.exists(outputs_grp['temp_file']):
            try:
                if not remove and g not in db_groups:
                    cleanup_mandb(None, outputs_grp, configs, logger)
                else:
                    # if we're overwriting
                    if remove:
                        logger.info(f"remove={remove}: removing {outputs_grp['temp_file']}")
                    # if found in database already
                    elif g in db_groups:
                        logger.info(f"{outputs_grp['temp_file']} found in db, removing")
                    os.remove(outputs_grp['temp_file'])
            except OSError as e:
                # remove if it can't be opened
                os.remove(outputs_grp['temp_file'])
    return error


def cleanup_obs(obs_id, policy_dir, errlog, configs, context=None,
                subdir='temp', remove=False):
    """
    For a given obs id, this function will search the policy_dir directory
    if it exists for any files with that obsnum in their filename. If any are
    found, it will run save_group_and_cleanup for that obs id.

    Arguments
    ---------
    obs_id: str
        Obs id to check and clean up
    policy_dir: str
        Directory to temp per-group output files
    errlog: fpath
        Filepath to error logging file.
    configs: fpath or dict
        Filepath or dictionary containing the preprocess configuration file.
    context: core.Context
        Optional. Context object used for data loading/querying.
    subdir: str
        Optional. Subdirectory to save the output files into.
    remove: bool
        Optional. Default is False. Whether to remove a file if found.
        Used when ``overwrite`` is True in driving functions.
    """

    if os.path.exists(policy_dir):
        found = False
        for f in os.listdir(policy_dir):
            if obs_id in f:
                found = True
                break

        if found:
            error = save_group_and_cleanup(obs_id, configs, context,
                                           subdir=subdir, remove=remove)
            if error is not None:
                f = open(errlog, 'a')
                f.write(f'\n{time.time()}, cleanup error\n{error[0]}\n{error[2]}\n')
                f.close()


def preproc_or_load_group(obs_id, configs_init, dets, configs_proc=None,
                          logger=None, overwrite=False):
    """
    This function is expected to receive a single obs_id, and dets dictionary.
    The dets dictionary must match the grouping specified in the preprocess
    config files. It accepts either one or two config strings or dicts representing
    an initial and a dependent pipeline stage. If the preprocess database entry for
    this obsid-dets group already exists then this function will just load back the
    processed tod calling either the ``load_and_preprocess`` or
    ``multilayer_load_and_preprocess`` functions. If the db entry does not exist or
    the overwrite flag is set to True then the full preprocessing steps defined in
    the configs are run and the outputs are written to a unique h5 file. Any errors,
    the info to populate the database, the file path of the h5 file, and the process
    tod are returned from this function. This function is expected to be run in
    conjunction with the ``cleanup_mandb`` function which consumes all of the outputs
    (except the processed tod), writes to the database, and moves the multiple h5 files
    into fewer h5 files (each <= 10 GB).

    Arguments
    ---------
    obs_id: str
        Obs id to process or load
    configs_init: fpath or dict
        Filepath or dictionary containing the preprocess configuration file.
    dets: dict
        Dictionary specifying which detectors/wafers to load see ``Context.obsdb.get_obs``.
    configs_proc: fpath or dict
        Filepath or dictionary containing a dependent preprocess configuration file.
    logger: PythonLogger
        Optional. Logger object or None will generate a new one.
    overwrite: bool
        Optional. Whether or not to overwrite existing entries in the preprocess manifest db.

    Returns
    -------
    error: str
        String indicating if the function succeeded in its execution or failed.
        If ``None`` then it succeeded in processing and the mandB should be updated.
        If ``'load_success'`` then axis manager was successfully loaded from existing preproc db.
        If any other string then processing failed and output will be logged in the error log.
    output_init: list
        Varies depending on the value of ``error``.
        If ``error == None`` then output is the info needed to update the manifest db.
        If ``error == 'load_success'`` then output is just ``[obs_id, dets]``.
        If ``error`` is anything else then output stores what to save in the error log.
    output_proc: list:
        See output_init for possible values.
    aman: Core.AxisManager
        Processed axis manager only returned if ``error`` is ``None`` or ``'load_success'``.
    """

    if logger is None:
        logger = init_logger("preprocess")

    error = None

    if type(configs_init) == str:
        configs_init = yaml.safe_load(open(configs_init, "r"))

    context_init = core.Context(configs_init["context_file"])

    if configs_proc is not None:
        if type(configs_proc) == str:
            configs_proc = yaml.safe_load(open(configs_proc, "r"))
        context_proc = core.Context(configs_proc["context_file"])

        group_by, groups, error = get_groups(obs_id, configs_proc, context_proc)
    else:
        group_by, groups, error = get_groups(obs_id, configs_init, context_init)

    if error is not None:
        return error[0], [error[1], error[2]], [error[1], error[2]], None

    all_groups = groups.copy()
    cur_groups = [list(np.fromiter(dets.values(), dtype='<U32'))]
    for g in all_groups:
        if g not in cur_groups:
            groups.remove(g)

        if len(groups) == 0:
            logger.warning(f"group_list:{cur_groups} contains no overlap with "
                           f"groups in observation: {obs_id}:{cur_groups}. "
                           f"No analysis to run.")
            error = 'no_group_overlap'
            return error, [obs_id, dets], [obs_id, dets], None

    db_init_exist = find_db(obs_id, configs_init, dets, context_init,
                            logger=logger)

    db_proc_exist = False
    if configs_proc is not None:
        db_proc_exist = find_db(obs_id, configs_proc, dets, context_proc,
                                logger=logger)

    if (not db_init_exist) and db_proc_exist and (not overwrite):
        logger.info('dependent db requires initial db if not overwriting')
        error = 'no_init_db'
        return error, [obs_id, dets], [obs_id, dets], None

    if db_init_exist and (not overwrite):
        if db_proc_exist:
            try:
                logger.info(f"both db and depdendent db exist for {obs_id} {dets} loading data and applying preprocessing.")
                aman = multilayer_load_and_preprocess(obs_id=obs_id, dets=dets, configs_init=configs_init,
                                                      configs_proc=configs_proc, logger=logger)
                error = 'load_success'
                return error, [obs_id, dets], [obs_id, dets], aman
            except Exception as e:
                error = f'Failed to load: {obs_id} {dets}'
                errmsg = f'{type(e)}: {e}'
                tb = ''.join(traceback.format_tb(e.__traceback__))
                logger.info(f"{error}\n{errmsg}\n{tb}")
                return error, [errmsg, tb], [errmsg, tb], None
        else:
            try:
                logger.info(f"init db exists for {obs_id} {dets} loading data and applying preprocessing.")
                aman = load_and_preprocess(obs_id=obs_id, dets=dets, configs=configs_init,
                                           context=context_init, logger=logger)
            except Exception as e:
                error = f'Failed to load: {obs_id} {dets}'
                errmsg = f'{type(e)}: {e}'
                tb = ''.join(traceback.format_tb(e.__traceback__))
                logger.info(f"{error}\n{errmsg}\n{tb}")
                return error, [errmsg, tb], [errmsg, tb], None

            if configs_proc is None:
                error = 'load_success'
                return error, [obs_id, dets], [obs_id, dets], aman
            else:
                try:
                    outputs_proc = save_group(obs_id, configs_proc, dets, context_proc, subdir='temp_proc')
                    init_fields = aman.preprocess._fields.copy()
                    init_fields.pop('valid_data', None)
                    logger.info(f"Generating new dependent preproc db entry for {obs_id} {dets}")
                    # pipeline for init config
                    pipe_init = Pipeline(configs_init["process_pipe"], plot_dir=configs_init["plot_dir"], logger=logger)
                    aman_cfgs_ref = get_pcfg_check_aman(pipe_init)
                    # pipeline for processing config
                    pipe_proc = Pipeline(configs_proc["process_pipe"], plot_dir=configs_proc["plot_dir"], logger=logger)

                    # tags from context proc
                    tags_proc = np.array(context_proc.obsdb.get(aman.obs_info.obs_id, tags=True)['tags'])

                    if "tags" in aman._fields:
                        aman.move("tags", None)
                    aman.wrap('tags', tags_proc)

                    proc_aman, success = pipe_proc.run(aman)

                    # remove fields found in aman.preprocess from proc_aman
                    for fld_init in init_fields:
                        if fld_init in proc_aman:
                            proc_aman.move(fld_init, None)

                    proc_aman.wrap('pcfg_ref', aman_cfgs_ref)

                except Exception as e:
                    error = f'Failed to run dependent processing pipeline: {obs_id} {dets}'
                    errmsg = f'Dependent pipeline failed with {type(e)}: {e}'
                    tb = ''.join(traceback.format_tb(e.__traceback__))
                    logger.info(f"{error}\n{errmsg}\n{tb}")
                    return error, [errmsg, tb], [errmsg, tb], None
                if success != 'end':
                    # If a single group fails we don't log anywhere just mis an entry in the db.
                    return success, [obs_id, dets], [obs_id, dets], None

                logger.info(f"Saving data to {outputs_proc['temp_file']}:{outputs_proc['db_data']['dataset']}")
                proc_aman.save(outputs_proc['temp_file'], outputs_proc['db_data']['dataset'], overwrite)

                if 'valid_data' in aman.preprocess:
                    aman.preprocess.move('valid_data', None)
                aman.preprocess.merge(proc_aman)

                return error, [obs_id, dets], outputs_proc, aman
    else:
        # pipeline for init config
        logger.info(f"Generating new preproc db entry for {obs_id} {dets}")
        try:
            pipe_init = Pipeline(configs_init["process_pipe"], plot_dir=configs_init["plot_dir"], logger=logger)
            aman_cfgs_ref = get_pcfg_check_aman(pipe_init)
            outputs_init = save_group(obs_id, configs_init, dets, context_init, subdir='temp')
            aman = context_init.get_obs(obs_id, dets=dets)
            tags = np.array(context_init.obsdb.get(aman.obs_info.obs_id, tags=True)['tags'])
            aman.wrap('tags', tags)
            proc_aman, success = pipe_init.run(aman)
            aman.wrap('preprocess', proc_aman)
        except Exception as e:
            error = f'Failed to run initial pipeline: {obs_id} {dets}'
            errmsg = f'Initial pipeline failed with {type(e)}: {e}'
            tb = ''.join(traceback.format_tb(e.__traceback__))
            logger.info(f"{error}\n{errmsg}\n{tb}")
            return error, [errmsg, tb], [errmsg, tb], None
        if success != 'end':
            # If a single group fails we don't log anywhere just mis an entry in the db.
            return success, [obs_id, dets], [obs_id, dets], None

        logger.info(f"Saving data to {outputs_init['temp_file']}:{outputs_init['db_data']['dataset']}")
        proc_aman.save(outputs_init['temp_file'], outputs_init['db_data']['dataset'], overwrite)

        if configs_proc is None:
            return error, outputs_init, [obs_id, dets], aman
        else:
            try:
                outputs_proc = save_group(obs_id, configs_proc, dets, context_proc, subdir='temp_proc')
                init_fields = aman.preprocess._fields.copy()
                init_fields.pop('valid_data', None)
                logger.info(f"Generating new dependent preproc db entry for {obs_id} {dets}")
                # pipeline for processing config
                pipe_proc = Pipeline(configs_proc["process_pipe"], plot_dir=configs_proc["plot_dir"], logger=logger)
                 # tags from context proc
                tags_proc = np.array(context_proc.obsdb.get(aman.obs_info.obs_id, tags=True)['tags'])

                if "tags" in aman._fields:
                    aman.move("tags", None)
                aman.wrap('tags', tags_proc)

                proc_aman, success = pipe_proc.run(aman)

                # remove fields found in aman.preprocess from proc_aman
                for fld_init in init_fields:
                    if fld_init in proc_aman:
                        proc_aman.move(fld_init, None)

                proc_aman.wrap('pcfg_ref', aman_cfgs_ref)

            except Exception as e:
                error = f'Failed to run dependent processing pipeline: {obs_id} {dets}'
                errmsg = f'Dependent pipeline failed with {type(e)}: {e}'
                tb = ''.join(traceback.format_tb(e.__traceback__))
                logger.info(f"{error}\n{errmsg}\n{tb}")
                return error, [errmsg, tb], [errmsg, tb], None
            if success != 'end':
                # If a single group fails we don't log anywhere just mis an entry in the db.
                return success, [obs_id, dets], [obs_id, dets], None

            logger.info(f"Saving data to {outputs_proc['temp_file']}:{outputs_proc['db_data']['dataset']}")
            proc_aman.save(outputs_proc['temp_file'], outputs_proc['db_data']['dataset'], overwrite)

            if 'valid_data' in aman.preprocess:
                aman.preprocess.move('valid_data', None)
            aman.preprocess.merge(proc_aman)

            return error, outputs_init, outputs_proc, aman


def cleanup_mandb(error, outputs, configs, logger=None, overwrite=False):
    """
    Function to update the manifest db when data is collected from the
    ``preproc_or_load_group`` function. If used in an mpi framework this
    function is expected to be run from rank 0 after a ``comm.gather``.
    See the ``preproc_or_load_group`` docstring for the varying expected
    values of ``error`` and the associated ``outputs``. This function will
    either:

    1) Update the mandb sqlite file and move the h5 archive from its temporary
    location to its permanent path if error is ``None``.

    2) Return nothing if error is ``load_success``.

    3) Update the error log if error is anything else.

    Arguments
    ---------
    error : str
        Error message output form preprocessing functions
    outputs : dict
        Dictionary including entries for the temporary h5 filename
        ('temp_file') and the obs_id group metadata and db entry (db_data).
        See save_group for more info.
    configs : dict
        Preprocessing configuration dictionary
    logger : PythonLogger
        Optional.  Python logger.
    overwrite : bool
        Optional. Delete the entry in the archive file if it exists and
        replace it with the new entry.
    """

    if logger is None:
        logger = init_logger("preprocess")

    if error is None:
        # Expects archive policy filename to be <path>/<filename>.h5 and then this adds
        # <path>/<filename>_<xxx>.h5 where xxx is a number that increments up from 0 
        # whenever the file size exceeds 10 GB.
        nfile = 0
        folder = os.path.dirname(configs['archive']['policy']['filename'])
        basename = os.path.splitext(configs['archive']['policy']['filename'])[0]
        dest_file = basename + '_' + str(nfile).zfill(3) + '.h5'
        if os.path.isabs(folder) and not(os.path.exists(folder)):
            os.makedirs(folder)
        while os.path.exists(dest_file) and os.path.getsize(dest_file) > 10e9:
            nfile += 1
            dest_file = basename + '_' + str(nfile).zfill(3) + '.h5'
        group_by = [k.split(':')[-1] for k in outputs['db_data'].keys() if 'dets' in k]
        h5_path = os.path.relpath(dest_file,
                                  start=os.path.dirname(configs['archive']['index']))

        src_file = outputs['temp_file']

        logger.debug(f"Source file: {src_file}")
        logger.debug(f"Destination file: {dest_file}")

        with h5py.File(dest_file,'a') as f_dest:
            with h5py.File(src_file,'r') as f_src:
                for dts in f_src.keys():
                    logger.debug(f"\t{dts}")
                    # If the dataset or group already exists, delete it to overwrite
                    if overwrite and dts in f_dest:
                        del f_dest[dts]
                    f_src.copy(f_src[f'{dts}'], f_dest, f'{dts}')
                    for member in f_src[dts]:
                        logger.debug(f"\t{dts}/{member}")
                        if isinstance(f_src[f'{dts}/{member}'], h5py.Dataset):
                            f_src.copy(f_src[f'{dts}/{member}'], f_dest[f'{dts}'], f'{dts}/{member}')
        logger.info(f"Saving to database under {outputs['db_data']}")
        db = get_preprocess_db(configs, group_by, logger)
        if len(db.inspect(outputs['db_data'])) == 0:
            db.add_entry(outputs['db_data'], h5_path)
        # make sure we close the db each time
        db.conn.close()
        os.remove(src_file)
    elif error == 'load_success':
        return
    else:
        folder = os.path.dirname(configs['archive']['index'])
        if not(os.path.exists(folder)):
            os.makedirs(folder)
        errlog = os.path.join(folder, 'errlog.txt')
        f = open(errlog, 'a')
        f.write(f'{time.time()}, {error}\n')
        if outputs is not None:
            f.write(f'\t{outputs[0]}\n\t{outputs[1]}\n')
        f.close()


def get_pcfg_check_aman(pipe):
    """
    Given a preprocess pipeline class return an axis manager containing
    the ordered steps of the pipeline with all arguments for each step.

    Arguments
    ---------
    pipe: _Preprocess class
        Preprocess pipeline class from which to build the step argument axis
        manager.
    """

    pcfg_ref = core.AxisManager()
    for i, pp in enumerate(pipe):
        pcfg_ref.wrap(f'{i}_{pp.name}', core.AxisManager())
        for memb in inspect.getmembers(pp, lambda a:not(inspect.isroutine(a))):
            if not memb[0][0] == '_':
                if type(memb[1]) is dict:
                    pcfg_ref[f'{i}_{pp.name}'].wrap(memb[0], core.AxisManager())
                    for itm in memb[1].items():
                        pcfg_ref[f'{i}_{pp.name}'][memb[0]].wrap(itm[0], str(itm[1]))
                else:
                    pcfg_ref[f'{i}_{pp.name}'].wrap(memb[0], memb[1])
    return pcfg_ref


def _check_assignment_length(a, b):
    """
    Helper function to check if the set of assignments in axis manager ``a`` matches
    the length of assignments in axis manager ``b``.

    Arguments
    ---------
    a: AxisManager
        Primary axis manager to cross check assignments with.
    b: AxisManager
        Secondary axis manager to cross check assignments with
    """

    aa = np.fromiter(a._assignments.keys(), dtype='<U32')
    bb = np.fromiter(b._assignments.keys(), dtype='<U32')

    if len(aa) != len(bb):
        return False, None, None
    else:
        return True, aa, bb


def check_cfg_match(ref, loaded, logger=None):
    """
    Checks that the ``ref`` and ``loaded`` axis managers containing the ordered
    preprocess pipelines match one another.

    Arguments
    ---------
    ref : AxisManager
        Reference axis manager for cross checking
    loaded : AxisManager
        Loaded axis manager for cross checking.
    logger : PythonLogger
        Optional. Python logger object.
    """

    if logger is None:
        logger = init_logger("preprocess")
    check, ref_items, loaded_items = _check_assignment_length(ref, loaded)
    if check:
        for ri, li in zip (ref_items, loaded_items):
            if ri != li:
                logger.warning('Config check fails due to ordered pipeline element names not matching.')
                return False
            else:
                if type(ref[ri]) is core.AxisManager and type(loaded[li]) is core.AxisManager:
                    check_cfg_match(ref[ri], loaded[li], logger)
                elif ref[ri] == loaded[li]:
                    continue
                else:
                    logger.warning(f'Config check fails due to arguments of {li} not matching')
                    return False
        return True
    else:
        logger.warning('Config check fails due to pipeline list not being of equal length')
        return False
